# This file contains implementations of PLControllerConnector

from pl_nls import *

class PLControllerConnectorI:
  """ PLControllerConnector Interface
  Contains signatures and specs of the interface to PLControllerConnector
  """

  def registerControllerReachableCB(self, requester, user_data):
    """ Register object requester as being interested in controller reachable 
    events.

    requester must have handleControllerReachableEvent(controller_id, 
    user_data) method.

    Returns:
      callback_id - a positive id that can be used to unregister the callback, 
                    if successful. Otherwise returns a negative number
    """
    pass

  def registerControllerUnreachableCB(self, requester, user_data):
    """ Register object requester as being interested in link down events.

    requester must have handleControllerUnreachableEvent(controller_id, 
    user_data) method.

    Returns:
      callback_id - a positive id that can be used to unregister the callback, 
                    if successful. Otherwise returns a negative number
    """
    pass
    
  def registerPacketConsumerCB(self, consumer, packet_type, user_data):
    """ Register consumer as being interested in packets of type packet_type.

    consumer must have consumeControllerPacket(pkt, src_controller, type, 
    user_data) method.
    packet_type must be between 0 and 99. Only one consumer can be registered
    for any given packet type.
    
    Returns:
      callback_id - a positive id that can be used to unregister the callback, 
                    if successful. Otherwise returns a negative number
    """
    pass

  def unregisterCB(self, callback_id):
    """ Unregisters the callback identified by the callback_id

    Returns:
      True if unregistered successfully
      False if callback_id is invalid
    """
    pass


  def sendPacket(self, pkt, dst, type):
    """ Send packet pkt of type type to controller dst.

    Returns:
      True is packet is sent successfully and false otherwise (the neighbor 
      might be down or there might not be such a neighbor)
    """
    pass


class PLControllerConnectorOrpa:
  """ PLControllerConnector for One Router Per AS case
  """

  def __init__(self, rtr):
    self.rtr = rtr

    # register callbacks with nls
    self.rtr.nls.registerNeighborLinkUpCB(self.handleNeighborLinkUpEvent, 
                                          None)
    self.rtr.nls.registerNeighborLinkDownCB(self.handleNeighborLinkDownEvent, 
                                            None)
    self.rtr.nls.registerPacketConsumerCB(self.consumeNeighborPacket, 
                                          PLNlsI.CC_PKT_TYPE, None)
    # setup the callback mechanism
    self.crcb = {}
    self.cucb = {}
    self.pccb = {}

  def registerControllerReachableCB(self, requester, user_data):
    cbid = self.get_cb_id()
    self.crcb[cbid] = (requester, user_data)
    return cbid

  def registerControllerUnreachableCB(self, requester, user_data):
    cbid = self.get_cb_id()
    self.cucb[cbid] = (requester, user_data)
    return cbid
    
  def registerPacketConsumerCB(self, consumer, pkt_type, user_data):
    if self.rtr.opts.debug:
      assert pkt_type < 100 and pkt_type >= 0
      assert pkt_type not in self.pccb

    if pkt_type >= 100 or pkt_type < 0 or pkt_type in self.pccb:
      return -1

    cbid = self.get_cb_id()
    self.pccb[pkt_type] = (consumer, user_data, cbid)
    return cbid

  def unregisterCB(self, cbid):
    if cbid in self.crcb:
      del self.crcb[cbid]
      return True
    if cbid in self.cucb:
      del self.cucb[cbid]
      return True
    
    for (pkt_type, data) in self.pccb.iteritems():
      if data[2] == cbid:
        del self.pccb[pkt_type]
        return True
    
    return False


  def sendPacket(self, pkt, dst, type):
    type_str = "%02d:" % type
    return self.rtr.nls.sendPacket(type_str + pkt, dst, PLNlsI.CC_PKT_TYPE)

  def handleNeighborLinkUpEvent(self, nbr_id, data):
    # For ORPA link up event is controller reachable event. Notify all receivers
    # of the controller reachable event.
    for rcvr, data in self.crcb.itervalues():
        rcvr(nbr_id, data)

  def handleNeighborLinkDownEvent(self, nbr_id, data):
    # For ORPA link down event is controller unreachable event. Notify all 
    # receivers of the controller unreachable event.
    for rcvr, data in self.cucb.itervalues():
        rcvr(nbr_id, data)

  def consumeNeighborPacket(self, raw_pkt, src_id, ntype, udata):
    """ Receive packet from NLS and give it to the registered consumers
    """
    ps = raw_pkt.split(":", 1)
    
    if self.rtr.opts.debug:
      assert len(ps) == 2 and len(ps[0]) == 2
    
    # if the packet is invalid, ignore it
    if len(ps) != 2 or len(ps[0]) != 2:
      return

    # deliver the packet to the registered consumer if any
    tp = int(ps[0])
    pkt = ps[1]
    if tp in self.pccb:
      self.pccb[tp][0](pkt, src_id, tp, self.pccb[tp][1])

  def get_cb_id(self):
    def is_id_used(id):
      if id in self.crcb:
        return True
      if id in self.cucb:
        return True
      for (pkt_type, data) in self.pccb.iteritems():
        if data[2] == id:
          return True
      return False

    for i in xrange(0, 10000000):
      if not is_id_used(i):
        return i

    assert False, "More than 10M callbacks are registered simultaneously"
