# This file contain utility classes for network/connection related functionality

from twisted.internet.protocol import ClientFactory, ReconnectingClientFactory,\
    ServerFactory
from twisted.protocols.basic import NetstringReceiver
from twisted.python import log, failure
from twisted.internet import error

import random
import pdb


class PLPeerProtocol(NetstringReceiver):
  """ Pathlet Peer Protocol. 
  Objects of this class represent connections to neighboring routers.
  """
  
  connectionDone = failure.Failure(error.ConnectionDone())
  connectionDone.cleanFailure()

  def __init__(self, rtr):
    self.rtr = rtr

  def stopProducing(self):
    """ Called when the connection is lost.
    """
    pass

  def resumeProducing(self):
    self.factory.resumeProducing(self.peer_addr)

  def pauseProducing(self):
    self.factory.pauseProducing(self.peer_addr)

  def connectionMade(self):
    self.peer_addr = (self.transport.getPeer( ).host, 
                      self.transport.getPeer( ).port)
    
    if self.rtr.opts.show_connection_events:
      #print "Connected to " + str(self.peer_addr)
      pass

    # register ourselves as producers
    self.transport.registerProducer(self, True)

    self.factory.peerConnectionMade(self)

  def connectionLost(self, reason = connectionDone):
    """ Called by the framework whenever this connection is broken.
    """ 
    if self.rtr.opts.show_connection_events:
      #print "Connection Lost to " + str(self.peer_addr)
      pass

    self.factory.peerConnectionLost(self.peer_addr)

  def stringReceived(self, pkt):
    """ Callback invoked by the NetstringReceiver when a complete
    msg (represented as a string) is received. Type 00 is reserved
    for peer id messages
    """
    if len(pkt) > 2:
      if pkt[0:3] == "00:":
        self.factory.handlePeerIdPkt(self, pkt[3:])
      else:
        self.factory.consumeLinkPacket(self.peer_addr, pkt)


class PerClientReconnectingFactory(ClientFactory):
  """ PerClientReconnectingFactory provides basic functionality to
  start and stop auto-reconnecting on a per-client basis.

  Note that clients should call my resetDelay method after they have
  connected successfully.

  max_delay     - Maximum number of seconds between connection attempts.
  initial_delay - Delay for the first reconnection attempt.
  factor        - Multiplicitive factor by which the delay grows
  jitter        - Percentage of randomness to introduce into the delay
                  length to prevent stampeding.
  max_retries   - Maximum number of retries or None to represent infinite
  """
  max_delay = 5
  initial_delay = 1.0
  factor = 1.1
  jitter = 0.1
  max_retries = None

  delays = {}
  retries = {}
  call_ids = {}

  dont_reconnect = {} # peer_addr -> True. If peer_addr is present
  failed_links = {} # peer_addr -> True

  def retryConnecting(self, connector):
    """ Have this connector connect again, after a suitable delay. 
    
    If a delayed call is already scheduled, don't schedule a new one
    """
    assert connector is not None
    peer_addr = (connector.host, connector.port)

    if peer_addr in self.dont_reconnect:
      del self.dont_reconnect[peer_addr]

    if peer_addr in self.failed_links:
      return
    
    # print "XXX: Retrying to connect to", peer_addr

    if (self.call_ids.has_key(peer_addr) and 
        self.call_ids[peer_addr] is not None):
      return

    if self.retries.has_key(peer_addr):
      self.retries[peer_addr] += 1
    else:
      self.retries[peer_addr] = 1

    if (self.max_retries is not None and 
        self.retries[peer_addr] > self.max_retries):
      if self.noisy:
        log.msg("Abandoning %s after %d retries." %
                connector, self.retries[peer_addr])
      return

    if self.delays.has_key(peer_addr):
      self.delays[peer_addr] = min(self.delays[peer_addr] * self.factor, 
                                 self.max_delay)
    else:
      self.delays[peer_addr] = self.initial_delay

    if self.jitter:
      self.delays[peer_addr] = \
          random.normalvariate(self.delays[peer_addr],
                               self.delays[peer_addr] * self.jitter)

    if self.noisy:
      log.msg("%s will retry in %d seconds" % (connector, 
                                               self.delays[peer_addr],))

    def reconnector():
      self.call_ids[peer_addr] = None
      connector.connect()

    self.call_ids[peer_addr] = \
        self.rtr.reactor.callLater(self.delays[peer_addr], reconnector)

  def stopTryingToConnect(self, connector):
    """ Removes the scheduled call to connect to peer_addr, if any. 

    Note that the connection might have already happened when this function 
    is called (we are just not notified by the kernel yet). Thus, the PeerCon 
    should be prepared to handle an arriving connection after a call to 
    stopTrying
    """
   
    peer_addr = (connector.host, connector.port)
    if connector.state == "connecting":
      connector.stopConnecting()
    if (self.call_ids.has_key(peer_addr) and 
        self.call_ids[peer_addr] is not None):
      self.call_ids[peer_addr].cancel()
      self.call_ids[peer_addr] = None
    else:
      # print "XXX: Did not find the call_id to stop"
      pass
    self.dont_reconnect[peer_addr] = True
 
  def resetRetryDelay(self, peer_addr):
    """Call me after a successful connection to reset.
    
    I reset the delay and the retry counter.
    """
    # print "XXX: Reseting delay for", peer_addr
    if peer_addr in self.delays:
      self.delays[peer_addr] = self.initial_delay
      self.retries[peer_addr] = 0
    if peer_addr in self.call_ids:
      self.call_ids[peer_addr] = None

  def clientConnectionFailed(self, connector, reason):
    """ Called by the framework when a connection that we attempted 
    was not successful.
    """
    peer_addr = (connector.host, connector.port)
    if self.rtr.opts.show_connection_events:
      #print "Connection failed to (%s:%s)" % peer_addr
      pass
    if (peer_addr not in self.dont_reconnect and 
        peer_addr not in self.failed_links):
      # print "XXX: Connection failed, retrying to connect to", peer_addr
      self.retryConnecting(connector)
 	
  def clientConnectionLost(self, connector, reason):
    """ Called by the framework when a connection that we initiated 
    was lost after being established.
    """
    peer_addr = (connector.host, connector.port)
    if (peer_addr not in self.dont_reconnect and 
        peer_addr not in self.failed_links):
      # print "XXX: Connection lost, retrying to connect to", peer_addr
      self.retryConnecting(connector)

  def buildProtocol(self, addr):
    """ Create an instance of the PLPeerProtocol. Called by the framework.

    Params:
      addr - An object implementing twisted.internet.interfaces.IAddress.

    Returns:
      None
    """  
    p = self.protocol(self.rtr)
    p.factory = self
    return p
