# This file contains implementations of PLController
from pl_pl import *
from pl_irp import *
from pl_utils import *
import networkx as NX
import sys
import time
import socket


class PLGraph(NX.DiGraph):
  """ Usage:
  - First, create an empty graph.
  - Then, add edges using add_pedge and add_vedge. These function automatically
    select the shortest edge and keep only that one.
  - Find shortest paths in the graph.

  All added pls has to be flat.
  """

  def __init__(self, rtr, buildEdgeSet = False):
    self.rtr = rtr
    NX.DiGraph.__init__(self)
    self.pids = {} # {(lid1, lid2) -> pid}
    if buildEdgeSet:
      self.edge_set = set()
      self.as_set = set()
      # {directed_as_link or aid -> [pid_recs]}, pid_recs are for pids whose
      # first inter-AS link is the key or that are purely internal to aid.
      # pid_rec = (weight, pid)
      self.sorted_pids = {}
    else:
      self.edge_set = None
      self.as_set = None

  def edgeNoise(self, lid1, lid2):
    return (hash((lid1, lid2)) & 32767) / 8388608.0

  def get_min_pids(self):
    """ returns a set of pids made up of pids for each edge and AS. Does not
    include local ASes. pids are pick
    """
    min_pids = set()
    # for each value in sorted_pids, sort it and pick the right number of pids
    for k, val in self.sorted_pids.iteritems():
      val.sort()
      if isinstance(k, tuple):
        if len(val) > PLControllerI.PLS_PER_LINK:
          map(lambda x: min_pids.add(x[1]), val[0:PLControllerI.PLS_PER_LINK])
        else:
          map(lambda x: min_pids.add(x[1]), val)
      elif isinstance(k, int):
        if len(val) > PLControllerI.PLS_PER_AS:
          map(lambda x: min_pids.add(x[1]), val[0:PLControllerI.PLS_PER_AS])
        else:
          map(lambda x: min_pids.add(x[1]), val)
      else:
        assert False, "k should be tuple or int but it is %s" % (k)

    return min_pids

  def get_dst_pids_from_paths(self, paths):
    def getLinks(path):
      res = []
      for i in range(0, len(path) - 1):
        res.append((path[i], path[i+1]))
      return res

    # returns a set of pids that make up the paths (to destinations only)
    tree_pids = set()
    for dst, p in paths.iteritems():
      if dst[1] == 0:
        for link in getLinks(p):
          tree_pids.add(self.pids[link])
    return tree_pids

  def get_reachable_pids_from_root(self, root_lid):
    if self.has_node(root_lid):
      # get all the reachable vnodes
      r_vnodes = set(NX.dfs_preorder(self, root_lid))
      # get and return all the reachable pathlets. 
      r_pls = set()
      for ends, pid in self.pids.iteritems():
        if ends[0] in r_vnodes:
          r_pls.add(pid)
      return r_pls
    else:
      return set()

  def get_min_pids_in_reachable_graph(self, root_lid, tree_pids):
    # we don't add pid that cross an edge if we have already picked a pid
    # that crosses the edge
    min_pids = set()
    if self.has_node(root_lid):
      # get all the reachable vnodes
      r_vnodes = set(NX.dfs_preorder(self, root_lid))
      
      # sort reachable pathlets into as_link and aid buckets like sorted_pids
      lim_sorted_pids = {}
      for ends, pid in self.pids.iteritems():
        # if pid is reachable
        if ends[0] in r_vnodes:
          pl = self.rtr.gsp.apls[pid].pl
          lid1 = pl.seq[0]
          lid2 = pl.seq[-1]
          w = len(pl.seq) - 1 + self.edgeNoise(lid1, lid2)

          if pl.isInternal():
            if pid[0] not in lim_sorted_pids:
              lim_sorted_pids[pid[0]] = []
            lim_sorted_pids[pid[0]].append((w,pid))
          else:
            aslink = self.getFirstASLink(pl.seq)
            if aslink not in lim_sorted_pids:
              lim_sorted_pids[aslink] = []
            lim_sorted_pids[aslink].append((w,pid))

      
      # add at most PLControllerI.PLS_PER_LINK PLControllerI.PLS_PER_AS
      # if we have already added some pids in tree_pids, account for those
      for k, val in lim_sorted_pids.iteritems():
        val.sort()
        if isinstance(k, tuple):
          val_pids = map(lambda x: x[1], val)
          added_pid_set = tree_pids.intersection(val_pids)
          for valpid in val_pids:
            if len(added_pid_set) < PLControllerI.PLS_PER_LINK:
              added_pid_set.add(valpid)
            else:
              break
          min_pids.update(added_pid_set)          
        elif isinstance(k, int):
          if len(val) > PLControllerI.PLS_PER_AS:
            map(lambda x: min_pids.add(x[1]), val[0:PLControllerI.PLS_PER_AS])
          else:
            map(lambda x: min_pids.add(x[1]), val)
        else:
          assert False, "k should be tuple or int but it is %s 2" % (k)

    return min_pids

  def get_limited_graph_pids_from_root(self, root_lid):
    if self.has_node(root_lid):
      gpaths = NX.single_source_dijkstra_path(self, root_lid)
      tree_pids = self.get_dst_pids_from_paths(gpaths)
      min_pids = self.get_min_pids_in_reachable_graph(root_lid, tree_pids)
      min_pids.update(tree_pids)
      return min_pids
    else:
      return set()

  def get_tree_pids_from_root(self, root_lid):
    # returns a set of pids that build up a shortest path tree from root_lid 
    # to all known destinations
    if self.has_node(root_lid):
      gpaths = NX.single_source_dijkstra_path(self, root_lid)
      return self.get_dst_pids_from_paths(gpaths)
    else:
      return set()

  def getASesAndEdges(self, pl):
    """ Returns a list of inter-as edges and a list of ases that pl traverses. 
    Assumes pl is flat.
    """
    edges = []
    ases = []
    last_asn = pl.seq[0][0]
    ases.append(last_asn)
    for vn in pl.seq:
      if last_asn != vn[0]:
        if last_asn < vn[0]:
          edges.append((last_asn, vn[0]))
        else:
          edges.append((vn[0], last_asn))
        last_asn = vn[0]
        ases.append(last_asn)
    return (ases, edges)

  def getFirstASLink(self, vlids):
      # assumes that vlids is not from an internal pathlet
      home_as = vlids[0][0]
      for lid in vlids:
        if home_as != lid[0]:
          return (home_as, lid[0])
      assert False

  def add_pedge(self, pl):
    """ Adds an edge from the first vnode in the pathlet to the last vnodes 
    with weight equal to to the number of vlinks. If there is already an edge
    connecting the two vnodes, replaces it iff the new edge is shorter.
    pl has to be flat (the only non-flat pls are long local pls and we don't
    add them to the graph)
    """
    if len(pl.seq) < 2:
      return

    pid = pl.getPid(assumeFlatAndSave = True)
    lid1 = pl.seq[0]
    lid2 = pl.seq[-1]
    w = len(pl.seq) - 1 + self.edgeNoise(lid1, lid2)

    # update as_set, edge_set, sorted_pids
    if self.edge_set is not None:
      ases, edges = self.getASesAndEdges(pl)
      self.as_set.update(ases)
      self.edge_set.update(edges)

      if pl.isInternal():
        if pid[0] not in self.sorted_pids:
          self.sorted_pids[pid[0]] = []
        self.sorted_pids[pid[0]].append((w,pid))
      else:
        aslink = self.getFirstASLink(pl.seq)
        if aslink not in self.sorted_pids:
          self.sorted_pids[aslink] = []
        self.sorted_pids[aslink].append((w,pid))
        
    # add pl to the graph and remember the pid of the added edge.
    if not self.has_edge(lid1, lid2):
      self.add_edge(lid1, lid2, w)
      self.pids[(lid1, lid2)] = pid
    elif self[lid1][lid2] > w:
      # if the new pathlet is shorter, replace it with the old
      self.add_edge(lid1, lid2, w)
      self.pids[(lid1, lid2)] = pid

  def add_vedge(self, lid1, lid2, w):
    """
    """
    w += self.edgeNoise(lid1, lid2)
    if not self.has_edge(lid1, lid2):
      self.add_edge(lid1, lid2, w)
      self.pids[(lid1, lid2)] = None
    elif self[lid1][lid2] > w:
      # if the new virtual pathlet is shorter, replace it with the old
      self.add_edge(lid1, lid2, w)
      self.pids[(lid1, lid2)] = None


class PLControllerI:
  """ PLController Interface
  Base class that contain general functionality
  """
  PLS_PER_LINK = 1
  PLS_PER_AS = 10


class PLControllerLTOrpa(PLControllerI):
  """ PLController for One Router Per AS case with LT policies

  A more precise definition of LT policies is that we have the 5 internal
  pathlets and a pathlet for each vnode exposed to us. The latter pathlets
  are from the corresponding egress vnode to the exposed vnode.

  VNodes:
    0 - sink destination vnode
    1 - customer ingress
    2 - customer egress
    3 - peer ingress
    4 - peer egress
    5 - provider ingress
    6 - provider egress
    
  Pathlets:
    1 - 1 -> 2
    2 - 1 -> 4
    3 - 1 -> 6
    4 - 3 -> 2
    5 - 5 -> 2

  PLControllerLTOrpa State:
    rels   - {rid -> relation}, dict storing the relation for all of our 
             neighboring ASes. rid is our customer (0), peer (1) or 
             provider (2)
    as2nbr - {asid -> neighbor_rid}
  """

  def __init__(self, rtr):
    self.rtr = rtr
    self.rels = {}
    self.asrels = {}
    self.as2nbr = {}

    # read the configutation file, constructing the rels
    should_exist = self.readConfig()
    if not should_exist:
      return
      
  def initialize(self):
    """ Called by the router after all modules are created
    """
    # created and install 6 LT vnodes and 4 LT pathlets [Requires Orpa]
    # we assume that there is a single controller. In the general case,
    # there should be some division or designation of who constructs 
    # the vnodes and the pathlets
    self.existing_vnodes = []
    self.buildLT()

    # expose the ingress vnodes
    rel2vnode = [1, 3, 5]
    irid = self.rtr.id
    for rid, rel in self.rels.iteritems():
      self.rtr.vman.exposeVNodeInst((irid[0], irid[1], rel2vnode[rel]), rid)

    # register with vman for vnode reach/unreach events.
    self.rtr.vman.registerVNodeReachableCB(self.handleVNodeReachableEvent, 
                                           None)
    self.rtr.vman.registerVNodeUnreachableCB(self.handleVNodeUnreachableEvent, 
                                             None)
    # register with irp for router reachable/unreachable events.
    # We interpret them as AS reachable/unreachable events
    self.rtr.irp.registerRouterReachableCB(self.handleRouterReachableEvent,
                                           None)
    self.rtr.irp.registerRouterUnreachableCB(self.handleRouterUnreachableEvent,
                                             None)

  def isCustomer(self, aid):
    return aid in self.asrels and self.asrels[aid] == 0

  def isPeer(self, aid):
    return aid in self.asrels and self.asrels[aid] == 1

  def isProvider(self, aid):
    return aid in self.asrels and self.asrels[aid] == 2

  def aplsChanged(self, changed_pids = None):
    """ This function should be called by any code that changes apls due to
        external events (pl ann/wdr, vnode reach/unreach, local pl added, etc).
        The caller code is responsible for changing the rib_ins/lpls/apls and 
        should call this method after it has done everything specific to itself.

        updated_pids is a list of pids that were updated, or a None when all 
        pids are updated
    """

    # update long local pathlets, for LT this is a noop
    self.updateLongLocalPathlets()

    # update some stats
    if self.rtr.opts.collect_stats:
      self.rtr.stats.apls_changed_update()
    
    # update rrib_outs to reflect the changes in apls.
    changed = self.buildAndSetAllRribOuts(changed_pids)
    
    # Send updates to all connected peers
    for aid in changed:
      self.rtr.gsp.sendTo(aid)

  def updateLongLocalPathlets(self):
    return

  def buildAndSetAllRribOuts_simple(self, changed_pids = None):
    """ Make each rrib_out eaqual to all of apls.
        Returns a list of aids that are connected and whose rrib_out has changed

        self.lpls: {pid -> PLPathlet} 
        self.apls: {pid -> PLPath}
        self.rib_in: {pid -> {aid -> PLPath}}
        self.rib_out: {pid -> {aid -> PLPath}}
        self.rrib_out: {pid -> set(aid)} 
    """
    gsp = self.rtr.gsp
    # build the current set of connected peers
    cntd = set()
    for aid in self.as2nbr.iterkeys():
      if self.isPeerUp(aid):
        cntd.add(aid)

    apidset = set(gsp.apls.iterkeys())
    # changed = []
    # loop through all the rrib_out pids. If the pid is in apidset, set its
    # set to cntd, else set it to empty set. We are using the same set 
    # objects! So, no one should update rrib_out other than us.
    gsp.rrib_out = {}
    for pid in apidset:
      gsp.rrib_out[pid] = cntd
    return cntd

  def buildPLLTGraph(self):
    """ Build the pathlet graph containing all pathlets known to us. 
    """
    gsp = self.rtr.gsp
    home_aid = self.rtr.id[0]
    pg = PLGraph(self.rtr, buildEdgeSet = True)

    # add all known pathlets
    for pp in gsp.apls.itervalues():
      pg.add_pedge(pp.pl)
    return pg

  def buildAndSetAllRribOuts(self, changed_pids = None):
    """ Builds a tree rooted at the vnode that we are exposing to aid.
        Furthermore, for every inter-as link, adds PLS_PER_LINK more pathlets.
        Returns True iff rrib_out for aid has been updated. 
    """
    gsp = self.rtr.gsp
    home_aid = self.rtr.id[0]

    # clear the rrib_out and build the pl graph
    gsp.rrib_out = {}
    pg = self.buildPLLTGraph()

    # build the current sets of connected neighbors
    cntd = set()
    custs = set()
    provs = set()
    peers = set()
    for aid in self.as2nbr.iterkeys():
      if self.isPeerUp(aid):
        cntd.add(aid)
        if self.isCustomer(aid):
          custs.add(aid)
        elif self.isPeer(aid):
          peers.add(aid)
        elif self.isProvider(aid):
          provs.add(aid)
        else:
          assert False

    # each pid is advertised to a set of aids that is a union of all aids in
    # some classes. Construct all possible sets of aids.
    # 000 - empty set, 001 - custs, 010 - peers, 100 - provs
    # 000, 001, 010, 011, 100, 101, 110, 111
    aid_sets = [set(), custs, peers, peers | custs, 
                provs, provs | custs, provs | peers, provs | peers | custs]

    # these are the graph pids that should be sent to corresponding peers
    # to ensure reachability of all known to us destinations (except our own)
    if 0:
      cust_graph_pids = pg.get_reachable_pids_from_root((home_aid, 1))
      peer_graph_pids = pg.get_reachable_pids_from_root((home_aid, 3))
      prov_graph_pids = pg.get_reachable_pids_from_root((home_aid, 5))
    elif 0:
      cust_graph_pids = pg.get_limited_graph_pids_from_root((home_aid, 1))
      peer_graph_pids = pg.get_limited_graph_pids_from_root((home_aid, 3))
      prov_graph_pids = pg.get_limited_graph_pids_from_root((home_aid, 5))
    else:
      cust_graph_pids = pg.get_tree_pids_from_root((home_aid, 1))
      peer_graph_pids = pg.get_tree_pids_from_root((home_aid, 3))
      prov_graph_pids = pg.get_tree_pids_from_root((home_aid, 5))

    # these are the extra per-link/per-as pids that should be told to all peers
    #min_pids = pg.get_min_pids()

    # loop through all the pids that should be announced to anybody and
    # set their aid sets to the right ones
    all_pids = cust_graph_pids.copy()
    all_pids.update(peer_graph_pids)
    all_pids.update(prov_graph_pids)
    #all_pids.update(min_pids)
    rrib_out = gsp.rrib_out
    for pid in all_pids:
      set_ind = 0
      if pid in cust_graph_pids:
        set_ind += 1
      if pid in peer_graph_pids:
        set_ind += 2
      if pid in prov_graph_pids:
        set_ind += 4
      rrib_out[pid] = aid_sets[set_ind]

    if self.rtr.opts.show_rrib_out_updates:
      print self.rtr.gsp.getRribOutStr()

    return cntd

  def isPeerUp(self, aid):
    return self.rtr.nls.isNeighborReachable(self.as2nbr[aid])

  def clearRribOut(self, aid):
    gsp = self.rtr.gsp
    for aids in gsp.rrib_out.itervalues():
        aids.discard(aid)
              
  def handleVNodeReachableEvent(self, lid, udata):
    # we do something only for remote vnodes
    aid = self.rtr.id[0]
    if lid[0] != aid:
      # for each remote AS, we are advertised only a single vnode, the ingress
      # for our class. We construct a pathlet from our egress vnode to the one
      # we are exposed.
      rel2evnode = [(aid, 2), (aid, 4), (aid, 6)]
      rel = self.rels[self.as2nbr[lid[0]]]
      self.setupLTPathlet(rel2evnode[rel], lid,)

  def handleVNodeUnreachableEvent(self, lid, udata):
    # local vnodes cannot become unreachable
    aid = self.rtr.id[0]
    assert lid[0] != aid
    rmed_pls = self.rtr.gsp.remoteVNodeUnreachable(lid)
    
    self.removeLocalPls(rmed_pls)

  def removeLocalPls(self, rmed_pls):
    # uninstall fentries for rmed_pls
    for rmpl in rmed_pls:
      self.uninstallPathlet(rmpl)

    # rebuild rrib_out for all aids
    self.aplsChanged()
    
  def handleRouterReachableEvent(self, rid, udata):
    """ build and set the rrib_out towards the peer. Ask the gossiper to send
    these requests.
    """
    self.buildAndSetAllRribOuts()
    self.rtr.gsp.sendTo(rid[0])

  def handleRouterUnreachableEvent(self, rid, udata):
    """ When a peer goes down, the handleRouterUnreachable simply calls gsp's
        peerAsUnreachable, which deletes the rib_in from the peer, builds apls
        and calls aplsChanged.
    """
    # tell that peer is unreachable to the gossiper, so that it can empty
    # the peer's rib_in, rebuild apls, and send out the updates
    self.rtr.gsp.peerAsUnreachable(rid[0])

  def buildLT(self):
    """ Build the LT vnodes and pathlets. Every AS builds all vnodes and 
    pathlets even if it does not have some class.

    rels   - {rid -> relation}, dict storing the relation for all of our 
             neighboring ASes. rid is our customer (0), peer (1) or 
             provider (2)
    """
    vman = self.rtr.vman
    aid = self.rtr.id[0]
    rid = self.rtr.id[1]
    pids = []

    # crate vnodes for classes that have some representation
    existing_rels = set(self.rels.values())
    for i in range(0,7):
      if ((i == 0) or
          ((i == 1 or i == 2) and 0 in existing_rels) or
          ((i == 3 or i == 4) and 1 in existing_rels) or
          ((i == 5 or i == 6) and 2 in existing_rels)):
        self.existing_vnodes.append((aid, i))
        vman.createVNodeInst((aid, rid, i))
    
    # create pathlets
    if (aid, 2) in self.existing_vnodes:
      # if customer egress is present, add the destination pathlet from it
      pids.append(self.setupLTPathlet((aid, 2), (aid, 0)))
    else:
      # otherwise, add pathlets to destination from all ingress vnodes
      if (aid, 3) in self.existing_vnodes:
        pids.append(self.setupLTPathlet((aid, 3), (aid, 0)))
      if (aid, 5) in self.existing_vnodes:
        pids.append(self.setupLTPathlet((aid, 5), (aid, 0)))

    if (aid, 1) in self.existing_vnodes and (aid, 2) in self.existing_vnodes:
      pids.append(self.setupLTPathlet((aid, 1), (aid, 2)))
    if (aid, 1) in self.existing_vnodes and (aid, 4) in self.existing_vnodes:
      pids.append(self.setupLTPathlet((aid, 1), (aid, 4)))
    if (aid, 1) in self.existing_vnodes and (aid, 6) in self.existing_vnodes:
      pids.append(self.setupLTPathlet((aid, 1), (aid, 6)))
    if (aid, 3) in self.existing_vnodes and (aid, 2) in self.existing_vnodes:
      pids.append(self.setupLTPathlet((aid, 3), (aid, 2)))
    if (aid, 5) in self.existing_vnodes and (aid, 2) in self.existing_vnodes:
      pids.append(self.setupLTPathlet((aid, 5), (aid, 2)))

    self.rtr.gsp.updateRemotePls(changed_pids = pids, rebuild = False)

  def uninstallPathlet(self, pl):
    """ Removes fentries for local pl from local vnodes. pl has to be flat.
    """
    assert pl.isFlat()
    home_as = self.rtr.id[0]
    for lid in pl.seq:
      if lid[0] == home_as:
        self.rtr.vman.removeFEAtVNode(lid, pl.fid)
    if self.rtr.opts.collect_stats:
      self.rtr.stats.fe_changed_update()

  def setupLTPathlet(self, lid1, lid2):
    """ Returns the pid of the created pathlets. 
    """
    # construct pathlet
    fid = self.getLTFid(lid1, lid2)
    pl = PLPathlet(fid, [lid1, lid2])
    pid = pl.getPid(assumeFlatAndSave = True)

    # add pathlet to the gossiper
    self.rtr.gsp.addLocalPathlet(pl)

    # install forwarding entry
    entry = ([], lid2)
    self.rtr.vman.installFEAtVNode(lid1, fid, entry)
    if self.rtr.opts.collect_stats:
      self.rtr.stats.fe_changed_update()
    return pid
      
  def getLTFid(self, lid1, lid2):
    """ 
    """
    return self.rtr.gsp.getUnusedLocalFid() 
    #asid = self.rtr.id[0]
    #assert lid1[0] == asid, "The first vnode has to be local"
    #if lid2[0] == asid:
    #  # purely local pathlet, one of 8 possibilities
    #  if lid1[1] == 1 and lid2[1] == 2:
    #    return 1
    #  elif lid1[1] == 1 and lid2[1] == 4:
    #    return 2
    #  elif lid1[1] == 1 and lid2[1] == 6:
    #    return 3
    #  elif lid1[1] == 3 and lid2[1] == 2:
    #    return 4
    #  elif lid1[1] == 5 and lid2[1] == 2:
    #    return 5
    #  elif lid1[1] == 1 and lid2[1] == 0:
    #    # either this one of the next one can exist, but not both
    #    return 6
    #  elif lid1[1] == 2 and lid2[1] == 0:
    #    return 6
    #  elif lid1[1] == 3 and lid2[1] == 0:
    #    return 7
    #  elif lid1[1] == 5 and lid2[1] == 0:
    #    return 8
    #  else:
    #    assert False, "Local pl %d->%d shouldn't exist" % (lid1[1], lid2[1])
    #else:
    #  # the end vnode is remote. 
    #  # For debugging we use deterministic fids that assume that there is a
    #  # limit of 100 on the number of exposed vnodes and ASes
    #  return self.rtr.gsp.getUnusedLocalFid() 
    #  # return lid2[0]*10000 + lid2[1]*100 + lid1[1] # TODO: change for above

  def readConfig(self):
    """ Read the config file.

    Determine if self should exist. If no, set the ptr to self in rtr to None
    and return False to signal that the __init__ should exit 
    without continuing.  If self should exist, the method constructs rels and
    return True.
    """
    while True:
      try:
        fd = open(self.rtr.opts.conf, "r")
      except IOError:
        hname = socket.gethostname()
        sys.stderr.write(hname + " has too many open files, retyring in 5 s\n")
        time.sleep(5)
      else:
        break
    
    rid = self.rtr.id
    section = 0
    subsection = 0
    should_exist = False
    if self.rtr.opts.everyone_customer:
      trans = [0, 0, 0]
    else:
      trans = [2, 1, 0]

    for line in fd:
      line = line.strip()
      if len(line) == 0 or line[0] == '#':
        continue
      
      ps = line.split()
      if ps[0] == "section":
        if ps[1] == "ctrl":
          section = 1
        else:
          section = 0
        continue

      if ps[0] == "subsection":
        if ps[1] == "controller_instances":
          subsection = 1
        elif ps[1] == "policy":
          subsection = 2
        continue


      if section == 1 and subsection == 1:
        assert len(ps) == 3, "controller_instance subsection line" \
            " \"%s\" is malformed" % line
        if (int(ps[0]), int(ps[1])) == rid:
          should_exist = True

      if section == 1 and subsection == 2:
        rid1 = (int(ps[0]), int(ps[1]))
        rid2 = (int(ps[2]), int(ps[3]))
        if self.rtr.opts.everyone_customer:
          rel = 0
        else:
          rel = int(ps[4])
        if rid1 == rid:
          if (not self.rtr.opts.everyone_customer and
              not isinstance(self, PLControllerPVOrpa)):
            if rel == 1:
              rel = 0
          assert rid2 not in self.rels
          assert rid2[0] not in self.rels
          self.rels[rid2] = trans[rel]
          self.asrels[rid2[0]] = trans[rel]
          self.as2nbr[rid2[0]] = rid2
        elif rid2 == rid:
          if (not self.rtr.opts.everyone_customer and
              not isinstance(self, PLControllerPVOrpa)):
            if rel == 1:
              rel = 2
          assert rid1 not in self.rels
          assert rid1[0] not in self.rels
          self.rels[rid1] = rel
          self.asrels[rid1[0]] = rel
          self.as2nbr[rid1[0]] = rid1

    fd.close()
    if not should_exist:
      self.rtr.ctrl = None
      return False
    else:
      return True


class PLControllerPVOrpa(PLControllerLTOrpa):
  """ PLController for One Router Per AS case with path vector policies

  The internal vnodes and pathlets are exactly the same as with pure LT
  policies. Configuration is also the same in that we only need to know
  what are our neighbors and what is our relationship

  Important note: we assume that no AS directly exposes its destination vnode
  
  VNodes:
    0 - destination
    1 - customer ingress
    2 - customer egress
    3 - peer ingress
    4 - peer egress
    5 - provider ingress
    6 - provider egress
    
  Pathlets:
    1 - 1 -> 2
    2 - 1 -> 4
    3 - 1 -> 6
    4 - 3 -> 2
    5 - 5 -> 2
    6 - 1 -> 0
    7 - 3 -> 0
    8 - 5 -> 0

  PLControllerLTOrpa State:
    rels   - {rid -> relation}, dict storing the relation for all of our 
             neighboring ASes. rid is our customer (0), peer (1) or 
             provider (2)
    asrels - {aid -> relation}
    as2nbr - {asid -> neighbor_rid}
    llpls  - {dest_vnode -> [local_pid, [remote_pid_seq]]}. The destination
             vnodes is a vnodes with vnode_id of 0. remote_pid_seq is a 
             list of remote pids that make up the long local pathlet. 
             local_pids is a list of local_pids that go to dest_vnode
  """

  # __init__, initialize, setupLTPathlet, getLTFid, buildLT, aplsChanged,
  #  are the same

  def __init__(self, rtr):
    PLControllerLTOrpa.__init__(self, rtr)

    # long local pathlet storage.
    self.llpls = {} # {dest_vnode -> [local_pid, [remote_pid_seq]]}

  def buildPLPVGraph(self):
    """ Build the pathlet graph containing virtual and remote pathlets.
    The graph contains no pathlets that start at us and our destination vnode
    is not part of the graph. Therefore, any path to any destination starting
    at our egress vnode is necessarily loop-free, i.e. does not traverse us 
    again.
    """
    gsp = self.rtr.gsp
    vman = self.rtr.vman
    home_aid = self.rtr.id[0]
    rel2evnode = [(home_aid, 2), (home_aid, 4), (home_aid, 6)]
    pg = PLGraph(self.rtr)

    # add remote pathlets
    for pp in gsp.apls.itervalues():
      pl = pp.pl
      if pl.seq[0][0] != home_aid:
        pg.add_pedge(pl)

    # add virtual pathlets from our corresponding egress vnode to the
    # exposed pathlet from each neighbor. ervnodes: rrid -> set(lid)
    for rrid, vset in vman.ervnodes.iteritems():
      assert len(vset) <= 1, \
          "%s exposed %d vnodes to %d" % (str(rrid), len(vset), home_aid)
      for lid in vset:
        local_egress_vnode = rel2evnode[self.rels[rrid]]
        pg.add_vedge(local_egress_vnode, lid, 1)

    return pg

  def buildAndSetAllRribOuts(self, changed_pids = None):
    """ Returns a list of aids that are connected and whose rrib_out has 
        changed.
        rrib_out to a neighbor n contains all of our purely local pathlets
        and pathlets that the neighbor is able to use, i.e. roughly those that
        BGP's export filter would have allowed.
    """
    gsp = self.rtr.gsp
    cust_set = set()
    prov_set = set()
    home_aid = self.rtr.id[0]

    # build the current sets of connected neighbors
    cntd = set()
    custs = set()
    provs = set()
    peers = set()
    for aid in self.as2nbr.iterkeys():
      if self.isPeerUp(aid):
        cntd.add(aid)
        if self.isCustomer(aid):
          custs.add(aid)
        elif self.isPeer(aid):
          peers.add(aid)
        elif self.isProvider(aid):
          provs.add(aid)
        else:
          assert False

    # clear the rrib_out
    gsp.rrib_out = {}

    # handle the local pids
    for pid, pl in gsp.lpls.iteritems():
      if not pl.isFlat(): # a non-flat pl in lpls has to be long local
        if pl.seq[0] == (home_aid, 2):
          # the first vnode is customer egress, announce this pid to all
          gsp.rrib_out[pid] = cntd
        elif pl.seq[0] == (home_aid, 4) or pl.seq[0] == (home_aid, 6):
          # the first vnode is peer or provider egress, announce to customers
          gsp.rrib_out[pid] = custs
        else:
          assert False, "BUG: lpls contain an unexpected pathlet 1"
      else:
        # pl is purely local
        if pl.seq[0] == (home_aid, 1):
          gsp.rrib_out[pid] = custs
        elif pl.seq[0] == (home_aid, 3):
          gsp.rrib_out[pid] = peers
        elif pl.seq[0] == (home_aid, 5):
          gsp.rrib_out[pid] = provs
        elif pl.seq[0] == (home_aid, 2):
          gsp.rrib_out[pid] = cntd
        else:
          assert False, "BUG: lpls contain an unexpected pathlet 2"

    return cntd


  def updateLongLocalPathlets(self):
    """ Makes the current set of long local pathlets be consistent with the
    set of known remote pathlets. This is done as follows:
    1. We use remote pathlets from apls to find current best paths to each
       reachable destination.
    2. For each currently reachable destination: If current best path is the
       same as what we had, all the information associated with the llpl to
       the destination is already correct and we don't do anything. If the
       current best path is not the same as what we had, we update all the
       state associated with the llpl towards the destination. This state
       is: the entry for dst in llpls, two entries in lpls for the pids of the
       old and the new pathlet, two entries in apls for the pids of the old
       and the new pathlets, forwarding entries for the old and new pls
    3. For all destinations that were reachable before but not now, delete
       theirs llpls and associated state - entry in llpls, entry in lpls, 
       entry in apls
    """
    gsp = self.rtr.gsp
    vman = self.rtr.vman
    home_aid = self.rtr.id[0]
    pg = self.buildPLPVGraph()

    #print "In updateLongLocalPls: pg = ", pg.edges()
    # remember the destinations that were reachable. If any of them is not
    # currently reachable, we will delete it after the loop.
    old_dsts = set(self.llpls.iterkeys())

    # build shortest paths from each egress vnode (to implement full BGP, one
    # would find shortest paths from all egress vnodes - one for each neighbor).
    # all the paths start with a virtual pathlet and then all pathlets 
    # are remote
    if pg.has_node((home_aid, 2)):
      cust_paths = NX.single_source_dijkstra_path(pg, (home_aid, 2))
    else:
      cust_paths = {}
    if pg.has_node((home_aid, 4)):
      peer_paths = NX.single_source_dijkstra_path(pg, (home_aid, 4))
    else:
      peer_paths = {}
    if pg.has_node((home_aid, 6)):
      prov_paths = NX.single_source_dijkstra_path(pg, (home_aid, 6))
    else:
      prov_paths = {}

    # compute all the reachable destinations 
    reachable_dsts = set(cust_paths.iterkeys())
    reachable_dsts.update(peer_paths.iterkeys())
    reachable_dsts.update(prov_paths.iterkeys())
    reachable_dsts = set(filter(lambda x: x[1] == 0, reachable_dsts))

    # handle each reachable destination
    for dst in reachable_dsts:
      if dst in cust_paths:
        best_path = cust_paths[dst]
      elif dst in peer_paths:
        best_path = peer_paths[dst]
      elif dst in prov_paths:
        best_path = prov_paths[dst]
      else:
        assert False, "Reachable destination is unreachable - bug"
        
      best_pl = None
      remote_pids = self.getRemotePids(pg, best_path)
      if dst in old_dsts:
        # we had a path to dst before. Compare it to the current best path.
        if not self.isSameLongLocal(remote_pids, dst):
          # current best path is not the same as the old one 
          # construct a pathlet for the current best path 
          best_pl = self.buildLongLocalPathlet(best_path, remote_pids)
          # delete the state of the old best pathlet
          old_pid = self.llpls[dst][0]
          old_flat_pl = gsp.apls[old_pid].pl
          self.uninstallPathlet(old_flat_pl)
          del self.llpls[dst]
          del gsp.lpls[old_pid]
          del gsp.apls[old_pid]
      else:
        # we did not have a path to dst before, construct the current best pl
        best_pl = self.buildLongLocalPathlet(best_path, remote_pids)

      # at this point, if best_pl is not None, we just need to set up its state.
      # state of the old pl is removed if it existed. If the best_pl is None, 
      # we don't need to do anything.
      if best_pl is not None:
        # insert flat version of best_pl into apls
        best_fpl = best_pl.flatten()
        best_pid = best_fpl.getPid(assumeFlatAndSave = True)
        ppath = PLPath(best_fpl, 0, [home_aid])
        gsp.apls[best_pid] = ppath
  
        # insert the pathlet into lpls and llpls
        gsp.lpls[best_pid] = best_pl
        self.llpls[dst] = [best_pid, remote_pids]

        # install forwarding entries
        self.installLongLocalPathlet(best_pl, remote_pids)
      
    # handle previously reachable but now unreachable destinations
    lost_dsts = old_dsts - reachable_dsts
    for lostd in lost_dsts:
      # delete the state of the unreachable destination
      old_pid = self.llpls[lostd][0]
      old_flat_pl = gsp.apls[old_pid].pl
      self.uninstallPathlet(old_flat_pl)
      del self.llpls[lostd]
      del gsp.lpls[old_pid]
      del gsp.apls[old_pid]
        
  def isSameLongLocal(self, remote_pids, dst):
    """ Returns True iff remote_pids are the same pids as the ones in the
    current long local pathlet towards dst in self.llpls. Assumes that dst
    exists in llpls
    TODO: There is a possibility for oscillations here because there might be
    multiple shortest paths even though we reduced this probability by adding
    random noise to edge weights.

    llpls  - {dest_vnode -> [local_pid, [remote_pid_seq]]}
    """
    return self.llpls[dst][1] == remote_pids
  
  def buildLongLocalPathlet(self, path, remote_pids):
    """ Idempotent function that just constructs the long local pathlet
    from remote_pids it goes through and vnode path.
    Returns the constructed (non-flat) pathlet.
    We assume that there is a single local vnode in the beginning
    """
    gsp = self.rtr.gsp
    fid = gsp.getUnusedLocalFid()
    #print "building pathlet", path, "with fid", fid
    seq = [path[0]]
    for rpid in remote_pids:
      seq.append(gsp.apls[rpid].pl)
    return PLPathlet(fid, seq)

  def installLongLocalPathlet(self, pl, remote_pids):
    """ Install forwarding entries for pl. remote_pids is a list of remote_pids
    that should be insterted into the packet header when the packet leaves last
    local vnode,
    """
    lid1 = pl.seq[0]
    lid2 = pl.seq[1].seq[0]
    fids = map(lambda x: x[2], remote_pids)
    fids.reverse()
    entry = (fids, lid2)
    self.rtr.vman.installFEAtVNode(lid1, pl.fid, entry)

  def getRemotePids(self, pg, path):
    """ Idempotent. Returns a list of remote pids that make up the path. 
    path is a list of vnodes starting at some local egress vnode and going out.
    We assume that there is a single local vnode in the beginning
    """
    plen = len(path)
    pids = pg.pids
    rpids = []
    for i in xrange(1, plen - 1):
      pid = pids[(path[i], path[i+1])]
      assert pid is not None
      rpids.append(pid)
    return rpids
            
  def handleRouterReachableEvent(self, rid, udata):
    """ build and set the rrib_out towards the peer. Ask the gossiper to send
    these requests.
    """
    self.buildAndSetAllRribOuts()
    self.rtr.gsp.sendTo(rid[0])

  def handleRouterUnreachableEvent(self, rid, udata):
    # tell that peer is unreachable to the gossiper, so that it can empty
    # the peer's rib_in, rebuild apls, and send out the updates
    self.rtr.gsp.peerAsUnreachable(rid[0])
 
  def handleVNodeUnreachableEvent(self, lid, udata):
    # For the PV policies, we don't need to do anything when a vnode becomes
    # reachable.
    return

  def handleVNodeReachableEvent(self, lid, udata):
    return

 






















###########################################################################  
####################### recycled code        ##############################   
###########################################################################  
#    
#  def aplsChanged_old(self):
#    """ This function should be called by any code that changes apls due to
#        external events (pl ann/wdr, vnode reach/unreach, local pl added, etc).
#        The caller code is responsible for changing the rib_ins/lpls/apls and 
#        should call this method after it's done everything specific to itself.
#    """
#    # update some stats
#    if self.rtr.opts.collect_stats:
#      self.rtr.stats.apls_changed_update()
#    
#    # update rrib_outs to reflect the changes in apls.
#    # For all the peers that are down, clear their rrib_outs but don't include
#    # them in the set of updated rrib_outs
#    for aid in self.as2nbr.iterkeys():
#      if self.isPeerUp(aid):
#        # print "Peer", aid, " is up, rebuilding rrib_out"
#        self.buildAndSetRribOut(aid)
#      else:
#        # print "Peer", aid, " is down, clearing rrib_out"
#        self.clearRribOut(aid)
#        
#    # Send updates to all connected peers
#    for aid in self.as2nbr.iterkeys():
#      if self.isPeerUp(aid):
#        self.rtr.gsp.sendTo(aid)
#
#
#  def buildAndSetRribOut_old(self, aid):
#    """ Builds a tree rooted at the vnode that we are exposing to aid.
#        Furthermore, for every inter-as link, adds PLS_PER_LINK more pathlets.
#        Returns True iff rrib_out for aid has been updated. 
#    """
#    # clear the rrib_out for aid
#    gsp = self.rtr.gsp
#    for aids in gsp.rrib_out.itervalues():
#        aids.discard(aid)
#    # get the vnode we export to aid. If we did not export any vnode, don't
#    # send any pathlets, i.e. leave the rrib_out empty and return
#    if not self.as2nbr[aid] in self.rtr.vman.evnodes:
#      return
#    s = self.rtr.vman.evnodes[self.as2nbr[aid]]
#    assert len(s) == 1
#    explid = iter(s).next()
#    # build the graph
#    # nodes are vlid tuples, edges are integers (number of vnode-level links),
#    # only the shortest edge is added to the graph. The pid of the shortest 
#    # edge in the graph is stored in pids dict - {(vlid1, vlid2) -> pid}.
#    g = NX.DiGraph(weighted=True)
#    pids = {}
#    for pid, pp in gsp.apls.iteritems():
#      pl = pp.pl # pl has to be flat
#      if len(pl.seq) < 2:
#        continue
#
#      lid1 = pl.seq[0]
#      lid2 = pl.seq[-1]
#      w = len(pl.seq) - 1
#      if not g.has_edge(lid1, lid2):
#        g.add_edge(lid1, lid2, w)
#        pids[(lid1, lid2)] = pid
#      else:
#        # if the new pathlet is shorter, replace it with the old
#        if g[lid1][lid2] > w:
#          g.add_edge(lid1, lid2, w)
#          pids[(lid1, lid2)] = pid
#
#    # compute the shortest path tree and extract all the pids building it.
#    def getLinks(path):
#      res = []
#      for i in range(0, len(path) - 1):
#        res.append((path[i], path[i+1]))
#      return res
#        
#    paths = NX.single_source_dijkstra_path(g, explid)
#    tree_pids = set()
#    for p in paths.itervalues():
#      for link in getLinks(p):
#        tree_pids.add(pids[link])
#
#    def getFirstASLink(vlids):
#      # assumes that vlids is not from an internal pathlet
#      home_as = vlids[0][0]
#      for lid in vlids:
#        if home_as != lid[0]:
#          return (home_as, lid[0])
#      assert False
#
#    # loop through all the known pathlet and build epls: 
#    # {directedd_as_link or aid -> [pids with link as the first inter-AS link]}
#    # For aid keys, the list is the list of pids that are internal to aid
#    epls = {}
#    for pid, pp in gsp.apls.iteritems():
#      if pp.pl.isInternal():
#        if pid[0] not in epls:
#          epls[pid[0]] = []
#        epls[pid[0]].append(pid)
#      else:
#        aslink = getFirstASLink(pp.pl.seq)
#        if aslink not in epls:
#          epls[aslink] = []
#        epls[aslink].append(pid)
#            
#    # add at most self.PLS_PER_LINK pids from each value in epls to tree_pids
#    for plist in epls.itervalues():
#      if len(plist) > self.PLS_PER_LINK:
#        map(lambda x: tree_pids.add(x), plist[0:self.PLS_PER_LINK])
#      else:
#        map(lambda x: tree_pids.add(x), plist)
#    # install the tree_pids into rrib_out
#    for tp in tree_pids:
#      if tp not in gsp.rrib_out:
#        gsp.rrib_out[tp] = set([aid])
#      else:
#        gsp.rrib_out[tp].add(aid)
#
#    if self.rtr.opts.show_rrib_out_updates:
#      print self.rtr.gsp.getRribOutStr()
#
#    return True # TODO: change this to return True only when rrib_out changed
