from pl_irp import *
from pl_utils import *

import pickle
import time

class PLMraiTimer:
  """ Class representing an MRAI timer 
  """
  def __init__(self, rtr):
    self.__rtr = rtr
    self.__excbs = {} # {cbid -> func(expired_aid)}
    # aid -> next expiration time
    # MRAI is not set if the key (aid) is not present
    self.__mrai = {}

  def registerExpiredCB(self, cb):
    # get unused id
    for i in xrange(0, 1000000):
      if i not in self.__excbs:
        break
    if i == 999999:
      assert False, "More than 1M callbacks are registered to the mrai timer"
    self.__excbs[i] = cb
    return i

  def unregisterExpiredCB(self, cbid):
    if cbid in self.__excbs:
      del self.__excbs[cbid]

  def expires_at(self, aid):
    """ Returns the time of next mrai expiration for the peer aid
    """
    if not self.__mrai.has_key(aid):
      return -1
    return self.__mrai[aid]

  def expired(self, aid):
    """ Twisted scheduler lets the timer know that mrai expired for aid
    has occured by calling this function 
    """
    assert self.__mrai[aid] <= time.time()
    if self.__mrai.has_key(aid):
      del self.__mrai[aid]
      for cb in self.__excbs.itervalues():
        cb(aid)

  def set(self, aid):
    """ Sets the mrai timer and schedules the mrai expiration event if
    the timer has not already been set. Otherwise, does nothing.
    Returns the next expiration time.
    """
    if not self.__mrai.has_key(aid):
      val = int(6*(3 + random.random())/4.0); # TODO: set back to 15
      self.__mrai[aid] = time.time() + val
      self.__rtr.reactor.callLater(val, self.expired, aid)
    return self.__mrai[aid]


class PLAplsTimer:
  """ Class to schedule updates to apls
  """
  def __init__(self, rtr):
    self.__rtr = rtr
    self.__exp_time = -1
    # list storing parameters to call updateRemotePls with 
    self.call_params = None

  def expires_at(self):
    """ Returns the expiration time
    """
    return self.__exp_time

  def expired(self):
    """ Twisted scheduler lets the timer know that timer expired 
    """
    # it is important that we set expiration time to -1 before 
    # calling updateRemotePls
    self.__exp_time = -1
    if self.call_params is not None:
      self.__rtr.gsp.updateRemotePls(self.call_params[0], self.call_params[1])
      self.call_params = None

  def updateCallParams(self, changed_pids, rebuild):
    # do something only if expiration time is not -1, i. e. the timer is set
    if self.__exp_time != -1:
      if self.call_params is None:
        # this is the first call to updateCallParams, so save the params
        self.call_params = [changed_pids, rebuild]
      else:
        # handle changed_pids first
        # if either the existing or the new changed_pids is None, set to None
        if changed_pids is None or self.call_params[0] is None:
          self.call_params[0] = None
        else:
          # since none of changed_pids is None, both of them must be lists
          # append the new elements to the existing ones
          ss = set(self.call_params[0])
          ss.update(changed_pids)
          self.call_params[0] = list(ss)
        
        # handle rebuild
        # rebuild is simply the or of the new and existing
        self.call_params[1] |= rebuild

  def set(self):
    """ Sets the timer and schedules the expiration event if
    the timer has not already been set. Otherwise, does nothing.
    Returns the next expiration time.
    """
    if self.__exp_time == -1:
      val = int(4*(2 + random.random())/3.0) # set back to 10
      self.__exp_time = time.time() + val
      self.__rtr.reactor.callLater(val, self.expired)
      self.call_params = None
    return self.__exp_time


class PLGossiperOrpa:
  """ Pathlet Gossiper Class. The role of the gossiper is to process regular 
  gossiping messages. It does not register to any events other than receiving
  pathlet advertisements and withdrawals. The controller registers to events 
  and calls appropriate functions on the gossiper to influence its behaviour.
  In particular, the controller decides which pathlets should be advertised to
  which peers and tells about this decision to the gossiper, which in turn 
  executes it. 

  Thus, the controller is esentially a local decision maker. It does not send 
  any messages itself. Instead, it asks other modules (vman, gossiper) to 
  execute its decisions.

  The gossiper contains the mrai timer, which is hidden from the controller.

  """

  def __init__(self, rtr):
    """
    Params:
      nbr_ases - a static list of neighboring ASes (integer ids)
    """
    self.rtr = rtr
    self.mrai = PLMraiTimer(rtr)
    self.mrai.registerExpiredCB(self.handleMraiExpired)
    self.apls_timer = PLAplsTimer(self.rtr)
    # last_used_local_fid is set to 8 to leave space for the 8 LT fids, which
    # are not obtained using getUnusedLocalFid. All other fids should be 
    # obtained through this function. If you are to change the number of 
    # reserved fids, also change 8 in the getUnusedLocalFid function.
    self.last_used_local_fid = 0
    # A small optimization for picking fids quickly the first round
    self.fids_looped = False
    # {aid -> True/False} stores the sends that have been paused
    self.paused_sends = {}
    for aid in self.rtr.ctrl.as2nbr.iterkeys():
      self.paused_sends[aid] = False
    # {aid -> True/False} stores True is the current sending process has sent
    # an announcement. If there is no current process towards aid, there is no
    # ker aid.
    self.sent_anns = {}
    # {pid -> PLPathlet} storing local pathlets in their original non-flat 
    # state. lpls is always a subset of apls
    self.lpls = {} 
    # {pid -> PLPath} storing all (local and remote) pathlets flattened out
    # chosen from rib_ins. PLPath objects (as well as dist and aspath in them)
    # are not shared with rib_ins but are shared with rib_outs. apls and 
    # rib_out's aspaths in PLPaths contain the self asid as the first elem.
    # Pathlet objects are shared in the whole system because they're immutable.
    self.apls = {} 
    # {pid -> {aid -> PLPath}} contains just want we received from each peer
    self.rib_in = {}
    # a set of all path currently present in rib_in
    self.rib_in_set = set()
    # {pid -> {aid -> PLPath}} actual transmitted ppaths to each peer.
    # The corresponding dictionaries are not deleted when there are no 
    # entries remaining. They can be cleaned periodically using callLater.
    self.rib_out = {} 
    # {pid -> set(aid)} pathlets (more precisely, their ppaths) that ctrl 
    # requests us to sent. rib_out is syncronized to rrib_out
    self.rrib_out = {} 

    # register for gossiper msgs (pl adv/wdr)
    self.cbids_irp = []
    self.cbids_nls = []
    self.cbids_irp.append(self.rtr.irp.registerPacketConsumerCB(
        self.consumeGossipPkt, PLIrpI.GOSSIP_CTRL_PKT_TYPE, None))

    for rid in self.rtr.ctrl.as2nbr.itervalues():
      self.cbids_nls.append(self.rtr.nls.registerResumePauseCB(
          self.resumeProducing, self.pauseProducing, rid, None))

  def stopReceiving(self):
    for cbid in self.cbids_irp:
      self.rtr.irp.unregisterCB(cbid)
    for cbid in self.cbids_nls:
      self.rtr.nls.unregisterCB(cbid)

  def remoteVNodeUnreachable(self, lid):
    """ We don't need to call updateRemotePls because we are removing some local
        vnodes and can easily do it here.
    """
    # remove all the local pathlets that rely on the lid from lpls and apls
    rmed_pls = []
    for pid, pl in self.lpls.items():
      if pl.isLidDependent(lid):
        del self.lpls[pid]
        del self.apls[pid]
        rmed_pls.append(pl)
    if self.rtr.opts.show_apls_updates:
      print self.getAplsStr()
    if self.rtr.opts.show_lpls_updates:
      print self.getLplsStr()
    # return a list of removed pathlets so that the controller can remove 
    # forwarding entries thus uninstalling the pathlets
    return rmed_pls
    
  def peerAsUnreachable(self, aid):
    """ 
        self.lpls: {pid -> PLPathlet} 
        self.apls: {pid -> PLPath}
        self.rib_in: {pid -> {aid -> PLPath}}
        self.rib_out: {pid -> {aid -> PLPath}}
        self.rrib_out: {pid -> set(aid)} 
    """ 
    # delete all the pathlets we learned from aid and all pathlets we have
    # have told aid about
    for pid, srcs in self.rib_in.iteritems():
      if aid in srcs:
        self.rib_in_set.remove(srcs[aid])
        del srcs[aid]
    for pid, dsts in self.rib_out.iteritems():
      if aid in dsts:
        del dsts[aid]    

    if self.rtr.opts.show_rib_in_updates:
      print self.getRibInStr()
    if self.rtr.opts.show_rib_out_updates:
      print self.getRibOutStr()

    # rebuild apls
    self.updateRemotePls()
    
  def addLocalPathlet(self, pl):
    """ If a pathlet with the same pid as pl alreaddy exists, the old pathlet
    is overwritten
    """
    # add the pathlet to local pathlets dict
    fpl = pl.flatten()
    pid = fpl.getPid(assumeFlatAndSave = True)
    if pid in self.lpls and self.rtr.opts.debug:
      print "Adding local pathlet %s that already exists in lpls" % (str(pid),)
    self.lpls[pid] = pl

    # insert the pathlet into all pathlets dict
    ppath = PLPath(fpl, 0, [self.rtr.id[0]])
    if pid in self.apls and self.rtr.opts.debug:
      print "Adding local pathlet %s that already exists in apls" % (str(pid),)
    self.apls[pid] = ppath

    if self.rtr.opts.show_lpls_updates:
      print self.getLplsStr()

  def pauseProducing(self, rid, udata):
    """ Called by the NLS to notify us that send buffers towards aid are full.
    """
    aid = rid[0]
    self.paused_sends[aid] = True

  def resumeProducing(self, rid, udata):
    """ Called by the NLS to notify us that send buffers towards aid can accept
        more data.
    """
    # Call doSendTo only when the process was paused. If it is not paused,
    # it means that we terminated while it was paused, so no need to continue
    # sending.
    aid = rid[0]
    if self.paused_sends[aid]:
      self.paused_sends[aid] = False
      # if mrai is not set, call the doSendTo to send the remaining pathlets.
      if self.mrai.expires_at(aid) > 0:
        # mrai is set, call doSendTo with send_wdrs_only=True
        self.doSendTo(aid, True)
      else:
        # The mrai is not set. Call doSendTo and set mrai if it tells so.
        if self.doSendTo(aid, False):
          self.mrai.set(aid)

  def handleMraiExpired(self, aid):
    # when mrai expires, sending should not be paused
    self.sendTo(aid)

  def doSendTo(self, aid, send_wdrs_only):
    """ MRAI has already been taken care of, just send the difference between
        the rrib_out and rib_out.

        Returns True iff the caller should set the mrai, which is the case, when
        we sent all messages we needed to (i.e. we were not paused) and their 
        included at least one non-withdrawal

        Should not be called for disconnected aids.
    """
    def findAndSendWdr(aid, it, items_to_del):
      """ Returns True if found and sent a withdrawal. False, if no more wdrs
          left.
      """
      #for pid, s in self.rib_out.iteritems():
      for pid, s, in it:
        if (aid in s and 
            (pid not in self.rrib_out or aid not in self.rrib_out[pid])):
          # we have found a pid for which we should send a withdrawal,
          # delete it from the rib_out
          #del self.rib_out[pid][aid]
          assert (pid, aid) not in items_to_del
          items_to_del.add((pid, aid))
          #pdb.set_trace()
          
          # send the wdr
          msg = PLGMsg(PLGMsg.GOSSIP_WDR_MSG, pid, None)
          smsg = pickle.dumps(msg)
          self.rtr.irp.sendPacket(smsg, 
                                  self.rtr.ctrl.as2nbr[aid], 
                                  PLIrpI.GOSSIP_CTRL_PKT_TYPE)
          return True
      return False
    def findAndSendAnn(aid, it):
      """ Returns True if found and sent an announcement. False, if no more 
          announcements are left.
      """
      #for pid, aids in self.rrib_out.iteritems():
      for pid, aids in it:
        if aid in aids and (pid not in self.rib_out or 
                            aid not in self.rib_out[pid] or
                            not self.rib_out[pid][aid].equals(self.apls[pid])):
          # we have found a pid for which we might send an announcement if
          # the path for it does not contain aid - sender-side loop detection
          if aid not in self.apls[pid].aspath:
            # insert the ppath we are sending in rib_out
            if pid not in self.rib_out:
              self.rib_out[pid] = {}
            self.rib_out[pid][aid] = self.apls[pid]
          
            # send the announcement
            msg = PLGMsg(PLGMsg.GOSSIP_ANN_MSG, pid, self.apls[pid])
            smsg = pickle.dumps(msg)
            self.rtr.irp.sendPacket(smsg, 
                                    self.rtr.ctrl.as2nbr[aid], 
                                    PLIrpI.GOSSIP_CTRL_PKT_TYPE)
            return True
          else:
            # if aid is in the path and we sent this pid to aid before, 
            # we need to send a withdrawal
            if pid in self.rib_out and aid in self.rib_out[pid]:
              # delete the item from rib_out
              del self.rib_out[pid][aid]

              # send the wdr
              msg = PLGMsg(PLGMsg.GOSSIP_WDR_MSG, pid, None)
              smsg = pickle.dumps(msg)
              self.rtr.irp.sendPacket(smsg, 
                                      self.rtr.ctrl.as2nbr[aid], 
                                      PLIrpI.GOSSIP_CTRL_PKT_TYPE)
              return True
      return False


    assert self.rtr.ctrl.isPeerUp(aid)
    wdrs_left = True
    sent_everything = False
    sent_ann = False
    items_to_del = set()
    rrib_out_it = self.rrib_out.iteritems()
    rib_out_it = self.rib_out.iteritems()
    while not self.paused_sends[aid]:
      if send_wdrs_only:
        # Find and send a wdr, if any. If no wdrs are left to be sent, 
        # break from the loop
        if not findAndSendWdr(aid, rib_out_it, items_to_del):
          sent_everything = True
          break    
      else:
        # try to send a withdrawal
        if wdrs_left:
          if findAndSendWdr(aid, rib_out_it, items_to_del):
            continue
          else:
            wdrs_left = False
        if findAndSendAnn(aid, rrib_out_it):
          sent_ann = True
          continue
        else:
          sent_everything = True
          break

    # delete the items we sent wdrs for but did not delete from rib_out
    for i in items_to_del:
      del self.rib_out[i[0]][i[1]]

    # the caller should_set_mrai only if we have sent_everything and 
    # we sent an ann now or at some time before in the same sending process
    if sent_everything and (sent_ann or (aid in self.sent_anns and 
                                         self.sent_anns[aid])):
      should_set_mrai = True
    else:
      should_set_mrai = False

    # if current sending process has ended, delete its state. Else, update it
    if sent_everything:
      if aid in self.sent_anns:
        del self.sent_anns[aid]
    else:
      if aid in self.sent_anns:
        self.sent_anns[aid] |= sent_ann
      else:
        self.sent_anns[aid] = sent_ann

    if self.rtr.opts.show_rib_out_updates:
      print self.getRibOutStr()

    return should_set_mrai

  def sendTo(self, dst_aid):
    """ Each call to sendTo starts a new sending process.
        If aid is None, send for all peers. Takes care of the MRAI timer.
    """
    # set aids to a list of aids to which we should try sending
    if dst_aid == None:
      aids = []
      for aid in self.rtr.ctrl.as2nbr.iterkeys():
        if self.rtr.ctrl.isPeerUp(aid):
          aids.append(aid)
    else:
      aids = []
      if self.rtr.ctrl.isPeerUp(dst_aid):
        aids.append(dst_aid)
    # try sending to each aid in aids
    for aid in aids:
      if aid in self.sent_anns:
        # A sending process is in progress, and it should be paused
        assert self.paused_sends[aid]
        # If the current process sent an announcement, set mrai. We don't need 
        # to call doSendTo here because the process will be resumed and since
        # the mrai timer is set, resume function will call doSendTo with asking
        # to send withdrawals only. Then, when the mrai expires, we will send
        # the announcements
        if self.sent_anns[aid]:
          self.mrai.set(aid)
        else:
          # The current process did not send any announcements. Thus, we don't
          # need to set the mrai. The new process will seemlessly merge with
          # the current one.
          pass
      else:
        # There is no active sending process, so it should not be paused
        assert not self.paused_sends[aid]
        if self.mrai.expires_at(aid) > 0:
          # Mrai is set, send only withdrawals now.
          self.doSendTo(aid, True)
        else:
          # The mrai is not set. Call doSendTo and set mrai if it tells so.
          if self.doSendTo(aid, False):
            self.mrai.set(aid)

  def updateRemotePls(self, changed_pids = None, rebuild = True):
    """ choose the apls from rib_ins and lpls.
    If rebuild is False, apls are not actually rebuild, but aplsChanged might 
    be called.
    If changed_pids is not None, it should be a list of pids that underwent
    a change and whose entry in apls might change as well. 
    """

    # TODO: remove after debugged
    # check that rib_in and lpls have disjoined pid sets
    # lpids = set(self.lpls.keys())
    # rpids = set(self.rib_in.keys())
    # assert len(lpids & rpids) == 0
    if self.apls_timer.expires_at() > 0:
      self.apls_timer.updateCallParams(changed_pids, rebuild)
      return
    
    if rebuild:
      updated_pids = []
      if changed_pids is None:
        # rebuild all apls
        oldpids = set(self.apls.keys())
        # make sure all local pathlets are in apls. If apl has a pid of a local
        # pathlet we assume that the pathlet is the same and do nothing
        # TODO: we might not need to touch local pathlets here, but it does 
        #       not hurt
        for pid, pl in self.lpls.iteritems():
          if pid in self.apls:
            oldpids.remove(pid)
          else:
            fpl = pl.flatten()
            ppath = PLPath(fpl, 0, [self.rtr.id[0]])
            self.apls[pid] = ppath
            updated_pids.append(pid)
    
        # loop through the rib_ins and update the apls
        for pid in self.rib_in.iterkeys():
          if len(self.rib_in[pid]) > 0:
            if self.choosePLPath(pid):
              updated_pids.append(pid)
            oldpids.discard(pid)
        # delete the pids in apls that we did not update - those 
        # still in oldpids
        for pid in oldpids:
          del self.apls[pid]
          updated_pids.append(pid)
        # show updated apls, if requested
        if self.rtr.opts.show_apls_updates:
          print self.getAplsStr()
      else:
        # we are given a list of pids to look at.
        for cpid in changed_pids:
          # If the pid is in lpls, get it from there and put into apls.
          # Else if the pid is in rib_ins and its entry is non-empty, choose one
          # Else the pid is not present anywhere, delete it from apls if present
          # TODO: we might not need to touch local pathlets here, but it does 
          #       not hurt
          if cpid in self.lpls:
            if cpid not in self.apls:
              fpl = self.lpls[cpid].flatten()
              ppath = PLPath(fpl, 0, [self.rtr.id[0]])
              self.apls[cpid] = ppath
              updated_pids.append(cpid)
          elif cpid in self.rib_in and len(self.rib_in[cpid]) > 0:
            if self.choosePLPath(cpid):
              updated_pids.append(cpid)
          else:
            if cpid in self.apls:
              del self.apls[cpid]
              updated_pids.append(cpid)
    else:
      updated_pids = changed_pids
      
    self.rtr.ctrl.aplsChanged(updated_pids)
    self.apls_timer.set()

  def choosePLPath(self, pid):
    """ Update the pid entry in apls from rib_ins. Assumes that the pid is not
        local.
        rib_in[pid] must not be empty.
        Returns True if the entry for pid was changed, else returns False.
    """
    def isParentAvailable(child):
      parent = PLPath(child.pl, 0, child.aspath[1:])
      return parent in self.rib_in_set
    inps = self.rib_in[pid]
    # If the current apls entry's parent still exists in rib_in don't change it
    #if pid in self.apls:
    #  for pp in inps.itervalues():
    #    if pp.isParent(self.apls[pid], assumeSamePid = True):
    #      return
    if pid in self.apls and isParentAvailable(self.apls[pid]):
      return False

    # If we are here, pid is either not in apls or rib_ins no longer have the 
    # parent of the current apls[pid]
    # Choose the plpath with the shortest aspath and the smallest source aid
    baid, bpp = inps.iteritems().next()
    for aid, pp in inps.iteritems():
      if (len(pp.aspath) < len(bpp.aspath) or
          len(pp.aspath) == len(bpp.aspath) and aid < baid):
        baid = aid
        bpp = pp
    # put the best path into apls
    naspath = [self.rtr.id[0]]
    naspath.extend(bpp.aspath)
    npp = PLPath(bpp.pl, 0, naspath)
    self.apls[pid] = npp
    return True

  def consumeGossipPkt(self, pkt, src_rid, type, udata):
    """
        self.lpls: {pid -> PLPathlet} 
        self.apls: {pid -> PLPath}
        self.rib_in: {pid -> {aid -> PLPath}}
        self.rib_out: {pid -> {aid -> PLPath}}
        self.rrib_out: {pid -> set(aid)} 

    self.type = type
    self.pid = pid
    self.ppath = ppath
    """
    gm = pickle.loads(pkt)
    src_aid = src_rid[0]
    home_aid = self.rtr.id[0]
    if gm.type == PLGMsg.GOSSIP_ANN_MSG:
      if self.rtr.opts.collect_stats:
        self.rtr.stats.gsp_adv_msg_count += 1
        #print "%d: %d -> %d: Ann %d-%s" % (self.rtr.stats.gsp_adv_msg_count,
        #            src_aid, home_aid, gm.pid[2], str(gm.ppath.pl.seq))
      # handle announcement packet. 
      # we ignore the announcement completely if the pathlet is local
      if gm.pid not in self.lpls:
        # the aspath does not have loops, put the pl in rib_in (either replacing
        # the old path or placing this path as the first from src_aid)
        if (home_aid not in gm.ppath.aspath and 
            len(filter(lambda x: x[0] == home_aid, gm.ppath.pl.seq)) == 0):
          if gm.pid not in self.rib_in:
            self.rib_in[gm.pid] = {}
          # update history
          if src_aid in self.rib_in[gm.pid]:
            self.rib_in_set.remove(self.rib_in[gm.pid][src_aid])
            self.rib_in_set.add(gm.ppath)
          else:
            self.rib_in_set.add(gm.ppath)
          # update rib_in
          self.rib_in[gm.pid][src_aid] = gm.ppath
        else:
          # If the path has a loop (either propagation loop or loop in pl.seq) 
          # delete the path src_aid sent us before, if any
          if gm.pid in self.rib_in and src_aid in self.rib_in[gm.pid]:
            del self.rib_in[gm.pid][src_aid]
    elif gm.type == PLGMsg.GOSSIP_WDR_MSG:
      if self.rtr.opts.collect_stats:
        self.rtr.stats.gsp_wdr_msg_count += 1
        #print "%d: %d -> %d: Wdr %s" % (self.rtr.stats.gsp_wdr_msg_count,
        #                           src_aid, home_aid, str(gm.pid))

      # handle withdrawal packet. Remove the ppath if any from the rib_in.
      if gm.pid in self.rib_in and src_aid in self.rib_in[gm.pid]:
        self.rib_in_set.remove(self.rib_in[gm.pid][src_aid])
        del self.rib_in[gm.pid][src_aid]
      if 0:
        # This code is no longer necessary because we always rebuild lpls 
        # from scratch
        # search through the lpls to find any pathlets that depended on the
        # withdrawn pathlet
        pls_to_rm = {}
        for pid, pl in self.lpls.iteritems():
          if pl.isPidDependent(gm.pid):
            pls_to_rm[pid] = pl
        # remove them from storage and notify ctrl so that it uninstalls them
        if len(pls_to_rm) > 0:
          for pid, pl in pls_to_rm.iteritems():
            del self.lpls[pid]
            del self.apls[pid]
          self.rtr.ctrl.removeLocalPls(pls_to_rm.values())
    else:
      assert False

    if self.rtr.opts.show_rib_in_updates:
      print self.getRibInStr()

    # rebuild apls
    self.updateRemotePls(changed_pids = [gm.pid])

  def getRribOutStr(self, countOnly = False):
    strs = []
    count = 0
    for pid, aids in self.rrib_out.iteritems():
      if len(aids) != 0:
        count += 1
        s = "%s --> %s" % (str(pid),str(list(aids)))
        strs.append(s)
    if countOnly:
      ns = str(count)
    else:
      strs.insert(0, "%s: %d's rrib_out totalling %d:" % \
                    (getLTime(), self.rtr.id[0], count))
      ns = "\n  ".join(strs)
    return ns

  def getRibOutStr(self, countOnly = False):
    strs = []
    count = 0
    for pid, aids in self.rib_out.iteritems():
      if len(aids) != 0:
        count += 1
        s = "%s ->> " % (str(pid),)
        for aid, pp in aids.iteritems():
          s += "#%d-%s " % (aid, pp.getLongStr())
        strs.append(s)
    if countOnly:
      ns = str(count)
    else:
      strs.insert(0, "%s: %d's rib_out totalling %d:" % \
                    (getLTime(), self.rtr.id[0], count))
      ns = "\n  ".join(strs)
    return ns

  def getRibInStr(self, countOnly = False):
    strs = []
    count = 0
    for pid, aids in self.rib_in.iteritems():
      if len(aids) != 0:
        count += 1
        s = "%s <<- " % (str(pid),)
        for aid, pp in aids.iteritems():
          s += "#%d-%s " % (aid, pp.getLongStr())
        strs.append(s)
    if countOnly:
      ns = str(count)
    else:
      strs.insert(0, "%s: %d's rib_in totalling %d:" % \
                    (getLTime(), self.rtr.id[0], count))
      ns = "\n  ".join(strs)
    return ns

  def getLplsStr(self, countOnly = False):
    strs = []
    count = 0
    i = 0
    for pl in self.lpls.itervalues():
      count += 1
      if i % 3 == 0:
        strs.append(pl.getLongStr())
      elif i % 3 == 1:
        strs[-1] += "\t" + pl.getLongStr()
      elif i % 3 == 2:
        strs[-1] += "\t" + pl.getLongStr()
      i += 1
    if countOnly:
      ns = str(count)
    else:
      strs.insert(0, "%s: %d's lpls totalling %d:" % \
                    (getLTime(), self.rtr.id[0], count))
      ns = "\n  ".join(strs)
    return ns

  def getAplsStr(self, countOnly = False):
    strs = []
    count = 0
    i = 0
    for pp in self.apls.itervalues():
      count += 1
      if i % 3 == 0:
        strs.append(pp.getLongStr())
      elif i % 3 == 1:
        strs[-1] += "\t" + pp.getLongStr()
      elif i % 3 == 2:
        strs[-1] += "\t" + pp.getLongStr()
      i += 1
    if countOnly:
      ns = str(count)
    else:
      strs.insert(0, "%s: %d's apls totalling %d:" % \
                    (getLTime(), self.rtr.id[0], count))
      ns = "\n  ".join(strs)
    return ns

  def getUnusedLocalFid(self):
    MAX_FIDS = 100000000

    if not self.fids_looped:
      if self.last_used_local_fid < MAX_FIDS:
        self.last_used_local_fid += 1
        return self.last_used_local_fid
      else:
        self.fids_looped = True
        self.last_used_local_fid = 0

    # extract all used fids:
    fids = set()
    for aid, vid, fid in self.lpls.iterkeys():
      fids.add(fid)
      
    # get the lowest unused one starting from last_used_local_fid
    for i in xrange(self.last_used_local_fid + 1, MAX_FIDS):
      if i not in fids:
        self.last_used_local_fid = i
        return i

    # if we got here, we did not find any unused fids in 
    # (self.last_used_local_fid, 100000000) range. Try from the beginning
    self.last_used_local_fid = 8
    for i in xrange(self.last_used_local_fid + 1, MAX_FIDS):
      if i not in fids:
        self.last_used_local_fid = i
        return i

    # if we got here, there are too many local fids, something is wrong
    assert False
