# This file contains implementations of PLIrp (Pathlet Intradomain Routing
# Protocol)

from pl_nls import *

class PLIrpI:
  """ PLIrp Interface
  Contains signatures and specs of the interface to PLIrp

  Pathlet IRP can deliver packets to any local and neighboring remote routers.
  That is the domain of the IRP. Router event callbacks apply to this domain, 
  i.e. IRP will only notify of events in this domain.
  """
  VMAN_CTRL_PKT_TYPE = 20
  DATA_PKT_TYPE = 21
  GOSSIP_CTRL_PKT_TYPE = 22

  def __init__(self):
    # setup the callback mechanism
    self.rrcb = {}
    self.rucb = {}
    self.pccb = {}

  def registerRouterReachableCB(self, requester, user_data):
    """ Register object requester as being interested in router reachable 
    events.

    requester must have handleRouterReachableEvent(router_id, user_data) method.

    Returns:
      callback_id - a positive id that can be used to unregister the callback, 
                    if successful. Otherwise returns a negative number
    """
    cbid = self.get_cb_id()
    self.rrcb[cbid] = (requester, user_data)
    return cbid

  def registerRouterUnreachableCB(self, requester, user_data):
    """ Register object requester as being interested in link down events.

    requester must have handleRouterUnreachableEvent(router_id, user_data) 
    method.

    Returns:
      callback_id - a positive id that can be used to unregister the callback, 
                    if successful. Otherwise returns a negative number
    """
    cbid = self.get_cb_id()
    self.rucb[cbid] = (requester, user_data)
    return cbid
    
  def registerPacketConsumerCB(self, consumer, pkt_type, user_data):
    """ Register consumer as being interested in packets of type packet_type.

    consumer must have consumeRouterPacket(pkt, src_router_id, type, user_data) 
    method.
    packet_type must be between 0 and 99. Only one consumer can be registered
    for any given packet type.
    
    Returns:
      callback_id - a positive id that can be used to unregister the callback, 
                    if successful. Otherwise returns a negative number
    """
    if self.rtr.opts.debug:
      assert pkt_type < 100 and pkt_type >= 0
      assert pkt_type not in self.pccb

    if pkt_type >= 100 or pkt_type < 0 or pkt_type in self.pccb:
      return -1

    cbid = self.get_cb_id()
    self.pccb[pkt_type] = (consumer, user_data, cbid)
    return cbid

  def unregisterCB(self, cbid):
    """ Unregisters the callback identified by the callback_id

    Returns:
      True if unregistered successfully
      False if cbid is invalid
    """
    if cbid in self.rrcb:
      del self.rrcb[cbid]
      return True
    if cbid in self.rucb:
      del self.rucb[cbid]
      return True
    
    for (pkt_type, data) in self.pccb.iteritems():
      if data[2] == cbid:
        del self.pccb[pkt_type]
        return True
    
    return False

  def sendPacket(self, pkt, dst, type):
    """ Send packet pkt of type type to router dst.

    Returns:
      True is packet is sent successfully and false otherwise (the router 
      might be down or there might not be such a router)
    """
    pass

  def isRouterReachable(self, rid):
    """ Check if the router in the domain is reachable or not. 

    Returns:
      True if rid is a domain router and is not known to be unreachable
      False if the rid is not in the domain or it is known to be unreachable
    """
    pass

  def get_cb_id(self):
    """ Internal method, not part of interface
    """
    def is_id_used(id):
      if id in self.rrcb:
        return True
      if id in self.rucb:
        return True
      for (pkt_type, data) in self.pccb.iteritems():
        if data[2] == id:
          return True
      return False

    for i in xrange(0, 10000000):
      if not is_id_used(i):
        return i

    assert False, "More than 10M callbacks are registered simultaneously"


class PLIrpOrpa(PLIrpI):
  """ PLIrp for One Router Per AS case
  """

  def __init__(self, rtr):
    # call the superclass's init method and set the rtr attribute
    PLIrpI.__init__(self)
    self.rtr = rtr
    
    # register callbacks with nls
    self.cbids_nls = []
    self.cbids_nls.append(self.rtr.nls.registerNeighborLinkUpCB(
        self.handleNeighborLinkUpEvent, None))
    self.cbids_nls.append(self.rtr.nls.registerNeighborLinkDownCB(
        self.handleNeighborLinkDownEvent, None))
    self.cbids_nls.append(self.rtr.nls.registerPacketConsumerCB(
        self.consumeNeighborPacket, PLNlsI.IRP_PKT_TYPE, None))

  def stopReceiving(self):
    for cbid in self.cbids_nls:
      self.rtr.nls.unregisterCB(cbid)

  def sendPacket(self, pkt, dst, type):
    type_str = "%02d:" % type
    return self.rtr.nls.sendPacket(type_str + pkt, dst,  PLNlsI.IRP_PKT_TYPE)

  def handleNeighborLinkUpEvent(self, nbr_id, data):
    # For ORPA link up event is router reachable event. Notify all receivers
    # of the router reachable event.
    for rcvr, data in self.rrcb.itervalues():
        rcvr(nbr_id, data)

  def handleNeighborLinkDownEvent(self, nbr_id, data):
    # For ORPA link down event is router unreachable event. Notify all receivers
    # of the router unreachable event.
    for rcvr, data in self.rucb.itervalues():
        rcvr(nbr_id, data)

  def consumeNeighborPacket(self, raw_pkt, src_id, ntype, udata):
    """ Receive packet from NLS and give it to the registered consumers
    """
    ps = raw_pkt.split(":", 1)
    
    if self.rtr.opts.debug:
      assert len(ps) == 2 and len(ps[0]) == 2
    
    # if the packet is invalid, ignore it
    if len(ps) != 2 or len(ps[0]) != 2:
      return

    # deliver the packet to the registered consumer if any
    tp = int(ps[0])
    pkt = ps[1]
    if self.rtr.opts.debug:
      #print "IRP of", self.rtr.id, "received pkt of type", tp
      pass
    if tp in self.pccb:
      self.pccb[tp][0](pkt, src_id, tp, self.pccb[tp][1])

  def isRouterReachable(self, rid):
    return self.rtr.nls.isNeighborReachable(rid)
