# This file contains implementations of PLNls (Pathlet Neighbor Link State)

from pl_con_utils import *
from twisted.internet.error import CannotListenError
import pdb
import time

class PLNlsI:
  """ PLNls Interface
  Contains signatures and specs of the interface to PLNls.
  """

  IRP_PKT_TYPE = 10
  CC_PKT_TYPE  = 11

  def __init__(self):
    # setup the callback mechanism
    self.lucb = {} # {cb_id -> (callable, udata)}
    self.ldcb = {} # {cb_id -> (callable, udata)}
    self.pccb = {} # {pkt_type  -> (consumer, user_data, cbid)}

    # Pause/Resume sending callbacks
    self.prcb = {} # {peer_rid -> [(callable_res, callable_pause, udata, cbid)]}

  def registerResumePauseCB(self, resumer, pauser, peer_rid, udata):
    """ Returns the cbid. Producer should accept two arguments: peer_rid, udata
    """
    cbid = self.get_cb_id()
    if peer_rid not in self.prcb:
      self.prcb[peer_rid] = []
    self.prcb[peer_rid].append((resumer, pauser, udata, cbid))
    return cbid

  def registerNeighborLinkUpCB(self, requester, user_data):
    """ Register requester function/method as being interested in link 
    up events.

    requester must take two arguments (neighbor, user_data). Suggested function
    name is handleNeighborLinkDownEvent

    Returns:
      callback_id - a positive id that can be used to unregister the callback, 
                    if successful.
    """
    cbid = self.get_cb_id()
    self.lucb[cbid] = (requester, user_data)
    return cbid

  def registerNeighborLinkDownCB(self, requester, user_data):
    """ Register requester as being interested in link down events.

    requester must take two arguments (neighbor, user_data). Suggested name for
    requester is handleNeighborLinkDownEvent

    Returns:
      callback_id - a positive id that can be used to unregister the callback, 
                    if successful.
    """
    cbid = self.get_cb_id()
    self.ldcb[cbid] = (requester, user_data)
    return cbid
    
  def registerPacketConsumerCB(self, consumer, pkt_type, user_data):
    """ Register consumer as being interested in packets of type packet_type.

    consumer must take 4 arguments (pkt, neighbor, type, user_data) 
    Suggested name consumeNeighborPacket
    packet_type must be between 10 and 99. Only one consumer can be registered
    for any given packet type.
    
    Returns:
      callback_id - a positive id that can be used to unregister the callback, 
                    if successful. Otherwise returns a negative number
    """
    if self.rtr.opts.debug:
      assert pkt_type < 100 and pkt_type >= 0
      assert pkt_type not in self.pccb

    if pkt_type >= 100 or pkt_type < 0 or pkt_type in self.pccb:
      return -1

    cbid = self.get_cb_id()
    self.pccb[pkt_type] = (consumer, user_data, cbid)
    return cbid

  def unregisterCB(self, cbid):
    """ Unregisters the callback identified by the callback_id

    Returns:
      True if unregistered successfully
      False if cbid is invalid
    """
    if cbid in self.lucb:
      del self.lucb[cbid]
      return True
    if cbid in self.ldcb:
      del self.ldcb[cbid]
      return True
    
    for pkt_type, data in self.pccb.iteritems():
      if data[2] == cbid:
        del self.pccb[pkt_type]
        return True

    # {peer_rid -> [(callable, callable, udata, cbid)]}
    for prid, clist in self.prcb.iteritems():
      for tp in clist:
        if tp[3] == cbid:
          clist.remove(tp)
          return True

    return False

  def sendPacket(self, pkt, dst, type):
    """ Send packet pkt of type type to neighbor dst.

    Returns:
      True is packet is sent successfully and false otherwise (the neighbor 
      might be down or there might not be such a neighbor)
    """
    pass

  def isNeighborReachable(self, rid):
    """ Check if the neighbor is reachable or not. 

    Returns:
      True if rid is a neighbor and is reachable
      False if rid is not a neighbor or not is reachable
    """
    pass

class PLNlsOrpa(PLNlsI, PerClientReconnectingFactory, ServerFactory):
  """ PLNls for One Router Per AS case

  PLNlsOrpa state:
    cntd_peers     - A dictionary from peer_ids to protocol objects.
    addr_to_peer   - A dictionary mapping (ip, port) to (as_id, router_id) 
                     for the currently established connections (independently
                     of who initiated the connection). cntd_peers 
                     and addr_to_peer should be always syncronized
    listen_addr    - A tuple (ip_addr, port) specifying which address the 
                     PeerCon should listen for connections, where ip_addr is a
                     doted string and port is an integer.
    peer_addrs     - A dictionary mapping peer_ids to the addresses they are
                     listening on: (as_id, router_id) -> (ip_addr, port)
    peer_connectors- {(as_id, router_id) -> [Connector object]}, static
    lucb/ldcp      - dictionaries {callback_id -> (receiver, user_data)} for
                     link up/link down events
    pccb           - dictionary {pkt_type -> (consumer, user_data, cb_id)}
                       
  """

  protocol = PLPeerProtocol

  def __init__(self, rtr):
    PLNlsI.__init__(self)
    self.rtr = rtr
    self.cntd_peers = {}
    self.addr_to_peer = {}
    self.peer_addrs = {}
    self.peer_connectors = {}

    self.readConfig()

    # start listening
    openedDataPort = False
    for i in range(0, 5):
      try:
        self.rtr.reactor.listenTCP(self.listen_addr[1]+i, self)
      except CannotListenError:
        print "port", self.listen_addr[1]+i, "was busy"
      else:
        openedDataPort = True
        break
    if not openedDataPort:
      print self.rtr.id[0], "could not open data port, exiting"
      self.rtr.shutDown()
      
    # fire connection requests. Lower id connects to higher id neighbors.
    for k in self.peer_addrs.iterkeys():
      if k > self.rtr.id:
        self.peer_connectors[k] = []
        for i in range(0, 5):
          cur_addr = (self.peer_addrs[k][0], self.peer_addrs[k][1] + i)
          # print "XXX: sending initial request to", cur_addr
          c = self.rtr.reactor.connectTCP(cur_addr[0], cur_addr[1], self)
          self.peer_connectors[k].append(c)

  def fail_link(self, peer_id):
    """ Fail the link to peer_id. If the link has already failed don't do 
    anything. The link might be just connecting, handle this case graciously.
    """
    # we can fail the link only if peer_id is higher than our id and peer_id
    # is indeed our peer
    if peer_id > self.rtr.id and peer_id in self.peer_addrs:
      print self.rtr.id[0], "failing link to", peer_id[0]
      for i in range(0,5):
        peer_addr = (self.peer_addrs[peer_id][0], 
                     self.peer_addrs[peer_id][1] + i)
        # print "XXX: fail_link: Stopping connections to", peer_addr
        self.failed_links[peer_addr] = True
        c = self.peer_connectors[peer_id][i]
        self.stopTryingToConnect(c)
        c.disconnect()
    else:
      print "requesting to fail link from a higher end or peer_id is invalid"

  def recover_link(self, peer_id):
    """ Bring up the link to peer_id. If the link is already up, don't do 
    anything.
    """
    if peer_id > self.rtr.id and peer_id in self.peer_addrs:
      # print "XXX: recovering link to", peer_id
      for i in range(0,5):
        peer_addr = (self.peer_connectors[peer_id][i].host,
                     self.peer_connectors[peer_id][i].port)
        del self.failed_links[peer_addr]
        self.retryConnecting(self.peer_connectors[peer_id][i])
    else:
      print "requesting to recover link from higher end or peer_id is invalid"

  def sendPacket(self, pkt, dst, type):
    if dst in self.cntd_peers and type < 100 and type >= 0:
      type_str = "%02d:" % type
      self.cntd_peers[dst].sendString(type_str + pkt)
      return True
    else:
      return False

  def isNeighborReachable(self, rid):
    return rid in self.cntd_peers

  def isNeighbor(self, rid):
    return rid in self.peer_addrs

  def consumeLinkPacket(self, peer_addr, raw_pkt):
    ps = raw_pkt.split(":", 1)
    
    if self.rtr.opts.debug:
      assert len(ps) == 2 and len(ps[0]) == 2
    
    # if the packet is invalid, ignore it
    if len(ps) != 2 or len(ps[0]) != 2:
      return

    # deliver the packet to the registered consumer if any
    tp = int(ps[0])
    pkt = ps[1]
    if self.rtr.opts.debug:
      #print "NLS of ", self.rtr.id, "received pkt of type", tp
      pass
    if tp in self.pccb:
      assert peer_addr in self.addr_to_peer
      self.pccb[tp][0](pkt, self.addr_to_peer[peer_addr], tp, 
                         self.pccb[tp][1])

  def resumeProducing(self, peer_addr):
    if peer_addr in self.addr_to_peer:
      rid = self.addr_to_peer[peer_addr]
      if rid in self.prcb:
        for tup in self.prcb[rid]:
          tup[0](rid, tup[2])
    else:
      if self.rtr.opts.debug:
        print "peer is not connected but we got a resumeProducing", peer_addr

  def pauseProducing(self, peer_addr):
    # {peer_rid -> [(resumer, pauser, udata, cbid)]}
    if peer_addr in self.addr_to_peer:
      rid = self.addr_to_peer[peer_addr]
      if rid in self.prcb:
        for tup in self.prcb[rid]:
          tup[1](rid, tup[2])
    else:
      if self.rtr.opts.debug:
        print "peer is not connected but we got a pauseProducing", peer_addr

  def peerConnectionMade(self, proto):
    """ The callback that is called by the peer protocol instance to
    notify the factory that the connection has been established. 

    We don't notify the registered receivers because we don't know whom have
    connected with. Once we receive an identification message from the other 
    side, we consider that the connection to the peer is established and 
    notify the registered receivers.

    Params:
      proto - PLPeerProtocol object representing the connection that has
              been established.
    """
    
    if self.rtr.opts.show_connection_events:
      #print self.rtr.id, "is sending id pkt to", proto.peer_addr
      pass

    # We cannot tell if we have initiated this connection, but
    # reset retry delay in any case.
    lps = set(self.peer_addrs.itervalues())
    found_base = False
    for i in range(0, 5):
      lpaddr = (proto.peer_addr[0], proto.peer_addr[1] - i)
      if lpaddr in lps:
        found_base = True
        break
    if found_base:
      for i in range(0,5):
        self.resetRetryDelay((lpaddr[0], lpaddr[1]+i))

    proto.sendString("00:%06d:%06d" % (self.rtr.id[0], self.rtr.id[1]))

  def peerConnectionLost(self, addr):
    """ The callback that is called by the peer protocol instance to
    notify the factory that the connection has been lost.

    Notify the interested receivers of this event. If the connection has been
    lost before we received an identification message, we don't do anything.
    
    Params:
      proto - PLPeerProtocol object representing the connection that has
              been lost

    """
    if self.addr_to_peer.has_key(addr):
      peer_id = self.addr_to_peer[addr]
      assert self.cntd_peers.has_key(peer_id)
      if self.rtr.opts.show_connection_events:
        print time.ctime(), self.rtr.id[0], "disconnected from", peer_id[0]

      # clear connection state. It is important that we delete the peer
      # before calling the callbacks
      del self.cntd_peers[peer_id]
      del self.addr_to_peer[addr]

      # nofity registered receivers
      for rcvr, data in self.ldcb.itervalues():
        rcvr(peer_id, data)
    
      # If we initiated this connection, retry connecting to all 
      # port the peer can be listening on
      if peer_id in self.peer_connectors:
        for c in self.peer_connectors[peer_id]:
          self.retryConnecting(c)      

  def handlePeerIdPkt(self, proto, msg):
    """ Processes the identification message received by proto.

    Identification pkt requires special handling because we don't know
    who the other side before receiving the pkt, which is needed for
    regular pkt dispatching.

    The identification message has format "asn:router_id" where both numbers 
    are six character (digit) ascii strings.

    Params:
      proto - Instance of PLPeerProtocol that received the message. The
              remote host's address can be retrieved using 
              proto.transport.getPeer().host proto.transport.getPeer().port
              The local address can be retrieved similarly using getHost().

      msg   - The identification msg represented as a string of format
              "asn,peer_id"

    Returns:
      None
    """
    ps = msg.split(":", 1)
    if self.rtr.opts.debug:
      assert len(ps) == 2 and len(ps[0]) == 6 and len(ps[1]) == 6
    
    # if the packet is invalid, ignore it
    if len(ps) != 2 or len(ps[0]) != 6 or len(ps[1]) != 6:
      return
    try:
      peer_id = (int(ps[0]), int(ps[1]))
    except ValueError:
      return

    if peer_id[0] == self.rtr.id[0]:
      return

    if self.rtr.opts.show_connection_events:
      #print self.rtr.id, "received ", peer_id, "'s id pkt."
      print time.ctime(), self.rtr.id[0], "connected to", peer_id[0]

    # If we requested this connection, stop trying to connect to the peer. 
    if peer_id in self.peer_connectors:
      for c in self.peer_connectors[peer_id]:
        self.stopTryingToConnect(c)

    # setup connection state
    assert peer_id not in self.cntd_peers
    assert proto.peer_addr not in self.addr_to_peer
    self.cntd_peers[peer_id] = proto
    self.addr_to_peer[proto.peer_addr] = peer_id

    # notify registered receivers of the peer up event, it is important we do
    # notification last
    for rcvr, data in self.lucb.itervalues():
        rcvr(peer_id, data)


  def readConfig(self):
    """ Read the config file fname and set listen_addr and peer_addrs
    """

    fd = open(self.rtr.opts.conf, "r")

    section = 0
    self.peer_addrs = {}
    
    for line in fd:
      line = line.strip()
      if len(line) == 0 or line[0] == '#':
        continue
      
      ps = line.split()
      if ps[0] == "section":
        if ps[1] == "physical_topology":
          section = 1
        elif ps[1] == "nls":
          section = 2
        else:
          section = 0
        continue

      if section == 1:
        ns = [int(x) for x in ps]
        assert len(ns) == 4, "physical_topology line %s is malformed" % line

        if (ns[0], ns[1]) == self.rtr.id:
          self.peer_addrs[(ns[2], ns[3])] = 1
        
        if (ns[2], ns[3]) == self.rtr.id:
          self.peer_addrs[(ns[0], ns[1])] = 1
        
      if section == 2:
        assert len(ps) == 4, "NLS line %s is malformed" % line

        peer_id = (int(ps[0]), int(ps[1]))
        addr = (ps[2], int(ps[3]))
        if peer_id == self.rtr.id:
          self.listen_addr = addr
        elif self.peer_addrs.has_key(peer_id):
          self.peer_addrs[peer_id] = addr
        
    assert self.listen_addr is not None
    assert not 1 in self.peer_addrs


  def get_cb_id(self):
    def is_id_used(id):
      if id in self.lucb:
        return True
      if id in self.ldcb:
        return True
      for (pkt_type, data) in self.pccb.iteritems():
        if data[2] == id:
          return True
      return False


    for i in xrange(0, 10000000):
      if not is_id_used(i):
        return i

    assert False, "More than 10M callbacks are registered simultaneously"
