class PLPathlet:
  """ Class to represent pathlets. This class is used for both hosted and
  remote pathlets. Pathlets are immutable.
  
  State:
    fid - an integer uniquely identifying this pathlet within the AS
    seq - list of tuples and/or references to other pathlets. The list is 
          ordered in the same way as the packet will travel. We set the
          convention that vnodes are repeated only when necessary, i.e.
          only when two subpathlets are next to each other in the list.
          Remote pathlets are flat, i.e. they contain only vnodes, but the
          same class is used for them as well. They are simply a special case.
  """

  def __init__(self, fid, seq = None):
    self.fid = fid
    if seq is None:
      self.seq = []
    else:
      self.seq = seq
    if not self.checkSeq():
      assert False, "Invalid pathlet"

  def isInternal(self):
    """ Returns True iff self is completely inside a single AS. Empty pathlet
    is considered internal. Assumes that self is flat.
    """
    #assert self.isFlat()
    if len(self.seq) == 0:
      return True
    home_as = self.seq[0][0]
    for lid in self.seq:
      if home_as != lid[0]:
        return False
    return True 

  def isLongLocal(self, aid):
    """ Returns True iff self starts locally but ends remotely. 
    aid is the local AS id.
    Assumes that self is flat and non-empty.
    """
    # assert self.isFlat()
    # assert len(self.seq) > 0
    if self.seq[0][0] == aid and self.seq[-1][0] != aid:
      return True
    else:
      return False

  def getShortStr(self):
    npl = self.flatten()
    return "%d:%s" % (npl.fid, str(npl.seq))

  def getLongStr(self):
    fpl = self.flatten()
    return "%d:%s" % (self.fid, str(fpl.seq))

  def isPidDependent(self, pid):
    """ Returns True iff self depends on pathlet pid
    """
    def isPidDepHelper(seq, pid):
      # return True iff pid is in seq
      if isinstance(seq, tuple):
        return False
      if seq.getPid() == pid:
        return True
      for el in seq.seq:
        if isPidDepHelper(el, pid):
          return True
      return False
    return isPidDepHelper(self, pid)

  def isLidDependent(self, lid):
    """ Returns True iff lid is in the self.seq
    """
    def isLidDepHelper(seq, lid):
      # returns True iff lid is in seq
      if isinstance(seq, tuple):
        return seq == lid
      for el in seq.seq:
        if isLidDepHelper(el, lid):
          return True
      return False
    return isLidDepHelper(self, lid)
    
  def flatten(self):
    """ Returns a new pathlet whose seq is a sequence of vnode tuples.
    """
    def flattenHelper(seq):
      # returns the flattened sequence without eliminating duplicates
      if isinstance(seq, tuple):
        return [seq]
      res = []
      for el in seq.seq:
        res.extend(flattenHelper(el))
      return res

    oseq = flattenHelper(self)
    
    # remove duplicates
    nseq = []
    for t in oseq:
      if len(nseq) == 0 or (len(nseq) > 0 and t != nseq[-1]):
        nseq.append(t)

    # create and return new flat pathlet
    return PLPathlet(self.fid, nseq)
    

  def isFlat(self):
    """ Returns True if the pathlet is flat, False otherwise
    """
    for i in self.seq:
      if not isinstance(i, tuple):
        return False
    return True

  def checkSeq(self):
    """ Check that the sequence is well-formed
    """
    def checkSeqHelper(el):
      if isinstance(el, PLPathlet):
        for ell in el.seq:
          if not checkSeqHelper(ell):
            return False
        return True
      elif isinstance(el, tuple):
        if len(el) == 2 and isinstance(el[0], int) and isinstance(el[1], int):
          return True
        else:
          return False
    return checkSeqHelper(self)

  def getPid(self, assumeFlatAndSave = False):
    if assumeFlatAndSave:
      ftup = self.seq[0]
    else:
      if not self.checkSeq() or len(self.seq) == 0:
        assert False, "Invalid pathlet"
      ftup = self.seq[0]
      while isinstance(ftup, PLPathlet):
        if len(ftup.seq) == 0:
          assert False, "Invalid pathlet"
        ftup = ftup.seq[0]
    return (ftup[0], ftup[1], self.fid)


class PLPathletPeerBase:
  """ Pathlet Peer Base is a class for storing pathlets learned from and 
  announced to the peer AS.
  
  PLPathletPeerBase State:
    ribi - {pid -> PLPathlet}, dict storing the pathlets we received from the 
           peer
    ribo - {pid -> PLPathlet}, dict storing the pathlets we sent to the
           peer
  """
  def __init__(self):
    self.ribi = {}
    self.ribo = {}

class PLPathletBase:
  """ Pathlet Base is a class that implements pathlet storage at a controller.
  (as_id, vnode_id, fid) is a unique id for a pathlet, which is abbreviated as
  a pid. 

  PathletBase State:
    lpls   - {pid -> PLPathlet}, dict storing local pathlets (pathlets created
             by our AS).
    krpls  - {pid -> PLPathlet}, dict storing all the known remote pathlets.
             It is the union of ribis in all the peers.
    peers  - {neighbor_as_id -> PLPathletPeerBase}, dict storing peer objects.
    pdep   - {pid -> set(PLPathlet)}, dict mapping the pid of a pathlet that is 
             used by all the local pathlets in the corresponding list.
    vdep   - {vlid -> set(PLPathlet)}, dict mapping domain vnodes to the set of
             local pathlet that go through this vnode.
  """
  def __init__(self, rtr):
    self.rtr = rtr
    self.lpls = {}
    self.peers = {}
    self.pdep = {}
    self.vdep = {}

  def store(self, pl):
    """ Store pathlet pl, which can be local as well as remote.

    store() updates the dependency maps if pl is local and ribi if pl is remote.
    """ 
    pass

  def remove(self, pid):
    """ Removes the pathlet pl.

    Recursively removes all the pathlets dependent on pl.

    Returns:
      List of pathlets that (directly or indirectly) depended on pl and were 
      removed.
    """

  def getDepOnVNode(self, vlid):
    """ Get a list of local pathlets that depend on a domain vnode vlid.

    getDepOnVNode is idempotent.
    """
    pass

  def getDepOnPathlet(self, pid):
    """ Get a list of local pathlets that depend on a local or remote pathlet
    pid.

    getDepOnPathlet is idempotent.
    """
    pass

  def clearPeer(self, pid):
    """ Clears the state learned from and sent to peer pid (int as_id). Removes
    all the pathlets depended on the peer (
    """
