# This file contains generally useful functionality

import time

def getLTime():
  t = time.localtime()
  return "%d:%d:%d" % (t[3], t[4], t[5])


class PLPath:
  """ Pathlet Path

  State:
    aspath - a list of ASes along which this advertisement came, the first
             element is the source of the message (not the source of the 
             pathlet).

  """
  def __init__(self, pl, dist, aspath):
    self.dist = dist      # always an integer
    self.pl = pl          # always a valid flat pathlet
    self.aspath = aspath  # always (possible empty) list

    # if dist is to be used, we need to modify __cmp__ and __hash__
    assert self.dist == 0
    #assert pl is not None
    #assert pl.isFlat()

  def __eq__(self, pp):
    return id(self.pl) == id(pp.pl) and self.aspath == pp.aspath
    
  def __hash__(self):
    h = 5381L
    for i in self.aspath:
      t = (h * 33) & 0xFFFFFFFFL
      h = t ^ i
    h = h ^ id(self.pl)
    return int(h % 0x7FFFFFFF)

  def isParent(self, pp, assumeSamePid = False):
    """ Returns True if pp was derived from self, i.e. if self.aspath is a 
    suffix of pp.aspath. Checks that pid is the same. Ignores the distance.
    """
    if not assumeSamePid:
      if self.pl.getPid() != pp.pl.getPid():
        return False
    if len(self.aspath) > len(pp.aspath):
      return False
    return self.aspath == pp.aspath[(len(pp.aspath) - len(self.aspath)):]

  def getShortStr(self):
    return "%s:%s" % (str(self.pl.getPid()), str(self.aspath))

  def getLongStr(self):
    return "%s:%s" % (self.pl.getShortStr(), str(self.aspath))

  def equals(self, pp):
    return self.aspath == pp.aspath

class PLGMsg:
  """ Pathlet Gossiper Message. 
  
  We use a single message class to send 2 types of messages. The type of the
  message is identified by the type field:

  Type 0: Pathlet Announcement
  Type 1: Pathlet Withdrawal

  PLCtrlMsg State:
    type   - 0/1 representing the type of the message
    pid    - full fid, or pathlet id, (as_id, vnode_id, fid)
    ppath  - flat pathlet being advertised. This pathlet has to be flat and
             must not have an aspath attribute
  """

  GOSSIP_ANN_MSG = 0
  GOSSIP_WDR_MSG = 1

  def __init__(self, type, pid, ppath):
    self.type = type
    self.pid = pid
    self.ppath = ppath

    assert len(self.pid) == 3
    if type == 0:
      assert self.ppath is not None
      assert self.ppath.pl is not None
      assert self.ppath.pl.isFlat(), "Only flat pathlets can be advertised"
    elif type == 1:
      assert ppath is None
    else:
      assert False
      


class PLDataMsg:
  """ Pathlet Data Message.

  Fields of the data msg:
    fids - a list of FIDs, each represented as an integer (which can be as
           big as we want since python's ints don't overflow). For performance
           reasons, the order of fids is reversed in the fids. The "first" fid
           to be processed by the next vnode is at the end of the list.
    dst  - lid of the destination vnode.
    ttl  - Time To Live as a small integer
    data - a string representing the data
    nh   - lid of the next hop vnode
    
  """

  def __init__(self, fids, dst, ttl, data, nh):
    """ Initialize the data message.

    Important: The list of fids will be modified by the handlers of the packet
               the instantiator of the packet has to provide a non-shared list
               and don't modify it after it was given to a packet. This warning
               does not apply to other params because they are immutable.
    """
    self.fids = fids
    self.dst = dst
    self.ttl = ttl
    self.data = data
    self.nh = nh

  def __str__(self):
    return "[DPKT dest=" +str(self.dst)+ " fids=" +str(self.fids)+ " nh=" + \
        str(self.nh)+ " data=" +self.data+ " ttl=" +str(self.ttl)+ "]"



