# This file contains implementations of PLVNodeManager

from pl_irp import *
from pl_vnode import *
import pickle

class PLVNodeManagerI:
  """ PLVNodeManager Interface
  Contains signatures and specs of the interface to PLVNodeManager

  The vnode manager does not keep track of what entries have been installed
  on what vnodes. It only keeps track of vnode reachability and executes
  controller commands.

  All vmanagers know about all the reachable vnodes. Thus, when a vnodes is 
  created or a vnode is exposed by a neighboring AS, all vmans need to be 
  updated. Vman instances (like controllers, routers, and links) are not 
  created dynamically in our implementation. This information is set once in 
  the configuration file. In practice, operators would update the configuration
  files and restart the router (or download the configuration file and do a 
  "soft restart" that will incorporate the new configuration file). Thus, at 
  startup all vmans know the location of all other vmans in the domain and 
  can send unicast/broadcast messages to all of them using the IRP. However,
  there is a distinction between local vmans and remote vmans. 
  """
  def __init__(self):
    # setup callback mechanism
    self.vircb = {}
    self.viucb = {}
    self.vrcb = {}
    self.vucb = {}

  def registerVNodeInstanceReachableCB(self, requester, user_data):
    """ Register object requester as being interested in vnode instance
    reachable events.

    requester must have handleVNodeInstanceReachableEvent(vnode_iid, user_data)
    method, where vnode_iid is the instance id - (as_id, router_id, vnode_id)

    Returns:
      callback_id - a positive id that can be used to unregister the callback, 
                    if successful. Otherwise returns a negative number
    """
    cbid = self.get_cb_id()
    self.vircb[cbid] = (requester, user_data)
    return cbid

  def registerVNodeInstanceUnreachableCB(self, requester, user_data):
    """ Register object requester as being interested in vnode instance
    unreachable events.

    requester must have handleVNodeInstanceUnreachableEvent(vnode_iid, 
    user_data) method.

    Returns:
      callback_id - a positive id that can be used to unregister the callback, 
                    if successful. Otherwise returns a negative number
    """
    cbid = self.get_cb_id()
    self.viucb[cbid] = (requester, user_data)
    return cbid

  def registerVNodeReachableCB(self, requester, user_data):
    """ Register object requester as being interested in vnode
    reachable events. A vnode is reachable when at least one instance of 
    the vnode is reachable.

    requester must have handleVNodeReachableEvent(vnode_lid, user_data)
    method, where vnode_lid is the logical vnode id - (as_id, vnode_id)

    Returns:
      callback_id - a positive id that can be used to unregister the callback, 
                    if successful. Otherwise returns a negative number
    """
    cbid = self.get_cb_id()
    self.vrcb[cbid] = (requester, user_data)
    return cbid

  def registerVNodeUnreachableCB(self, requester, user_data):
    """ Register object requester as being interested in vnode
    unreachable events. A vnode is unreachable if all of its instances are
    unreachable.

    requester must have handleVNodeUnreachableEvent(vnode_lid, user_data) 
    method.

    Returns:
      callback_id - a positive id that can be used to unregister the callback, 
                    if successful. Otherwise returns a negative number
    """
    cbid = self.get_cb_id()
    self.vucb[cbid] = (requester, user_data)
    return cbid


  def unregisterCB(self, cbid):
    """ Unregisters the callback identified by the callback_id

    Returns:
      True if unregistered successfully
      False if cbid is invalid
    """
    if cbid in self.vircb:
      del self.vircb[cbid]
      return True
    if cbid in self.viucb:
      del self.viucb[cbid]
      return True
    if cbid in self.vrcb:
      del self.vrcb[cbid]
      return True
    if cbid in self.vucb:
      del self.vucb[cbid]
      return True
    return False


  def installFEAtVNodeInst(self, vnode_iid, fid, entry):
    """ Install Forwarding Entry fid->entry at VNode Instance vnode_iid.

    Returns:
      True if successfully and false otherwise.
    """
    pass

  def removeFEAtVNodeInst(self, vnode_iid, fid):
    """ Remove Forwarding Entry for fid at VNode Instance vnode_iid.

    Returns:
      True if successfully and false otherwise.
    """
    pass

  def installFEAtVNode(self, vnode_lid, fid, entry):
    """ Install Forwarding Entry fid->entry at all instances of vnode_lid.

    It tries to install the entry on as many instances as possible.

    Returns:
      A list of vnode_iids on which the entry was not successfully installed.
    """
    pass

  def removeFEAtVNode(self, vnode_lid, fid):
    """ Remove Forwarding Entry for fid at all instances of vnode_lid.

    It tries to remove the entry from as many instances as possible.

    Returns:
      A list of vnode_iids on which the entry was not successfully removed.
    """
    pass

  def createVNodeInst(self, vnode_iid):
    """ Create a vnode instance vnode_iid

    Returns:
      True if the vnode inst vnode_iid can possibly be created successfully.
      False if vnode inst vnode_iid already exists.
    """
    pass

  def deleteVNodeInst(self, vnode_iid):
    """ Delete a vnode instance vnode_iid

    Returns:
      True if the vnode inst vnode_iid can possibly be deleted successfully or
           if it already does not exist
      False if vnode inst vnode_iid cannot be deleted.
    """
    pass

  def exposeVNodeInst(self, viid, rrid):
    """ Instruct the vman running on the router hosting viid to expose it to
    the remote router rrid.

    In debug mode, raises an exception if (1) the viid does not exist, (2) the
    rid does not exist, (3) there is no direct connection between the router
    hosting viid and rid (this is required because exposed vnodes are required
    to reside on the router that is directly connected to the neighboring AS)

    Returns:
      True if the (possibly) remote vman was sent the command successfully.
      False if the vman cannot be connected or if the arguments are invalid.
    """
    pass

  def getUnusedVNodeLid(self):
    """ Returns an lid that is not used in the home AS.
    """
    pass


  def get_cb_id(self):
    """ Not part of the interface. Returns an unused id for the callback
    """
    def is_id_used(id):
      if id in self.vircb:
        return True
      if id in self.viucb:
        return True
      if id in self.vrcb:
        return True
      if id in self.vucb:
        return True
      return False

    for i in xrange(0, 10000000):
      if not is_id_used(i):
        return i

    assert False, "More than 10M callbacks are registered simultaneously"


class PLVNodeManagerOrpa(PLVNodeManagerI):
  """ PLVNodeManager for One Router Per AS case

  From ther perspective of the vman there are "hosted vnodes" - vnodes that are
  residing on the same router at the vman, "local vnodes" - vnodes that are
  residing on routers belonging to the same as at the host router, "exposed 
  vnodes" - vnodes that are exposed to the home AS by the neighboring ASes.

  PLVNodeManagerOrpa State:
    vman_locs   - static list, [vman_hosting_router_id], storing ids of routers 
                  hosting vmans of interest. vmans of interest are vmans hosted
                  by local routers and directly connected remote AS routers.
                  For dynamic information we can always as the irp's 
                  isRouterReachable method. If the rotuer is reachable, we 
                  assume that the vman is reachable. This list is read from the
                  configuration file. It does not include self's rid.
    vnodes      - dictionary, lid -> PLVNode, storing all local vnodes.
    vnode_locs  - dictionary, lid -> set(rid), storing locations of all vnodes
                  in the domain (hosted, local, and exposed vnodes). This
                  dictionary is updated when we receive vnode exposure messages
                  and when a new local vnode is created. This dictionary stores
                  only currently reachable vnodes through currently reachable
                  locations. For example, if a router that hosted a vnode 
                  becomes unreachable, the vnode might still be reachable if 
                  there is another router that hosts it and is reachable.
    evnodes     - a dictionary, rrid -> set(vlid), storing vnodes that we are 
                  exposing, where rrid is the id of the remote router that has
                  to be self's direct neighbor, and [vlid] is a list of vnode
                  lids that are exposed to rrid. This mapping is updated through
                  calls to exposeVNodeInst by the controller. Whenever the 
                  mapping is updated, the diff (or the whole list if easier) is
                  sent to the neighbor (which then updates it vnode_locs and
                  communicates the change to all other local vmans)
    ervnodes    - a dictionary, rrid -> set(vlid), storing the remote vnodes
                  exposed to us by our direct neighbors rrid from remote ASes.
                  This dictionary always reflects the last state that was sent 
                  to us by the rrid. ervnodes are always included in the 
                  vnode_locs. The reason we need to have a separate dictionary
                  is because normally self owns only vnodes that reside on
                  self.rtr. ervnodes is the other set of vnodes that self owns
                  and needs to send to other routers when they become reachable
                  or when ervnodes change.
  """

  def __init__(self, rtr):
    # call the superclass's init method and set the rtr attribute
    PLVNodeManagerI.__init__(self)
    self.rtr = rtr

    # register callbacks with irp
    self.cbids_irp = []
    self.cbids_irp.append(self.rtr.irp.registerRouterReachableCB(
        self.handleRouterReachableEvent, None))
    self.cbids_irp.append(self.rtr.irp.registerRouterUnreachableCB(
        self.handleRouterUnreachableEvent, None))
    self.cbids_irp.append(self.rtr.irp.registerPacketConsumerCB(
        self.consumeRouterPacket, PLIrpI.VMAN_CTRL_PKT_TYPE, None))
    self.cbids_irp.append(self.rtr.irp.registerPacketConsumerCB(
        self.consumeRouterPacket, PLIrpI.DATA_PKT_TYPE, None))

    # vnode managers' locations [static]
    self.vman_locs = []

    # local vnodes and remote vnode locations
    # self.vnodes are changing only when a vnode is created/destroyed (either 
    # through a local or remote command)
    self.vnodes = {}

    # vnode_locs is always reflecting the locations of connected routers. 
    # vnode_locs' state is maintained through a simple broadcast mechanism.
    # The mechanism is based on the notion of the vnode ownership. Each vnode
    # instance is owned by the vman at the same router and remote exposed 
    # vnodes are owned by the direct neighbor of the exposing vnode. It is the
    # responsibility of the owner of the vnode (instance) to ensure that all 
    # local vmans know about its ownees. Whenever a router becomes unreachable,
    # all other routers delete the vnodes that resided on the router. When a 
    # router that has a vman becomes reachable to a set A of routers, each 
    # of the vmans in A, sends all the information it needs to send to the r 
    # and r's vman sends its info to all the vmans in A.
    self.vnode_locs = {}

    # vnodes we are exposing [dynamic in the sense that it changes at run time
    # and static in the sense that router reachable/unreachable events don't
    # affect it]
    # all values of self.evnodes are subsets of self.vnodes. 
    self.evnodes = {}

    # vnodes exposed to us
    self.ervnodes = {}
    
  def stopReceiving(self):
    for cbid in self.cbids_irp:
      self.rtr.irp.unregisterCB(cbid)

  def installFEAtVNodeInst(self, viid, fid, entry): # Requires Orpa
    """ Install Forwarding Entry fid->entry at VNode Instance vnode_iid.

    If fid key already exists, overwrites it (in debug mode, raises an error)

    Returns:
      True if successfully and false otherwise.
    """
    lid = (viid[0], viid[2])

    assert lid in self.vnodes

    if self.rtr.opts.debug:
      assert not self.vnodes[lid].containsFid(fid)

    self.vnodes[lid].installFE(fid, entry)
    return True

  def removeFEAtVNodeInst(self, viid, fid): # Requires Orpa
    """ Remove Forwarding Entry for fid at VNode Instance vnode_iid.

    Returns:
      True if successfully. Raises an error if fid does not exist or if there
      is no such vnode
    """
    lid = (viid[0], viid[2])
    assert lid in self.vnodes
    self.vnodes[lid].removeFE(fid)
    return True

  def installFEAtVNode(self, vlid, fid, entry): # Requires Orpa
    """ Install Forwarding Entry fid->entry at all instances of vnode_lid.

    It tries to install the entry on as many instances as possible.

    Returns:
      A list of vnode_iids on which the entry was not successfully installed.
    """
    assert vlid in self.vnodes

    if self.rtr.opts.debug:
      assert not self.vnodes[vlid].containsFid(fid)

    self.vnodes[vlid].installFE(fid, entry)
    return []

  def removeFEAtVNode(self, vlid, fid): # Requires Orpa
    """ Remove Forwarding Entry for fid at all instances of vnode_lid.

    It tries to remove the entry from as many instances as possible.

    Returns:
      A list of vnode_iids on which the entry was not successfully removed.
      Raises an error if there is no vnodes vlid
    """
    assert vlid in self.vnodes
    self.vnodes[vlid].removeFE(fid)
    return []

  def createVNodeInst(self, viid): # Requires Orpa
    """ Create a vnode instance vnode_iid

    Returns:
      True if the vnode inst vnode_iid can possibly be created successfully.
      False if vnode inst vnode_iid already exists.
    """
    lid = (viid[0], viid[2])
    rid = (viid[0], viid[1])

    if self.rtr.opts.debug:
      assert rid == self.rtr.id
      assert not lid in self.vnodes
      
    if rid != self.rtr.id or lid in self.vnodes:
      return False

    self.vnodes[lid] = PLVNode(self.rtr, viid)
    if lid not in self.vnode_locs:
      self.vnode_locs[lid] = set()
    self.vnode_locs[lid].add(rid)
    # TODO can be optimized
    self.sendToAllVMans()
    return True

  def deleteVNodeInst(self, viid): # Requires Orpa
    """ Delete a vnode instance vnode_iid

    Returns:
      True if the vnode inst vnode_iid can possibly be deleted successfully or
           if it already does not exist
      False if vnode inst vnode_iid cannot be deleted.
    """
    lid = (viid[0], viid[2])
    rid = (viid[0], viid[1])

    if self.rtr.opts.debug:
      assert rid == self.rtr.id
    
    if rid != self.rtr.id:
      return False
    if not lid in self.vnodes:
      return False

    del self.vnodes[lid]
    self.vnode_locs[lid].remove(rid)

    # TODO: can be optimized
    self.sendToAllVMans()
    return True
    
  def getUnusedVNodeLid(self):
    """ Returns an lid that is not used in the home AS.
    """
    asid = self.rtr.id[0]
    for i in xrange(1, 10000000):
      if not (asid, i) in self.vnodes:
        return (asid, i)
    assert False, "More than 10M vnodes are created in an AS"

  def exposeVNodeInst(self, viid, rrid): # Requires Orpa
    """ Expose viid to remote neighbor rrid

    Returns:
      True if viid is a hosted vnode and rrid is indeed a remote neighbor.
      False, otherwise (in debug mode raises an error)
    """
    
    vlid = (viid[0], viid[2])
    if self.rtr.opts.debug:
      assert self.rtr.nls.isNeighbor(rrid)
      assert not self.rtr.isLocal(rrid)
      assert vlid in self.vnodes

    if (not self.rtr.nls.isNeighbor(rrid) or self.rtr.isLocal(rrid) or
        not vlid in self.vnodes):
      return False
    
    if rrid not in self.evnodes:
      self.evnodes[rrid] = set()

    self.evnodes[rrid].add(vlid)
    self.sendToVMan(rrid)
    return True
    
  def handleRouterReachableEvent(self, rid, udata):
    """ Process the router reachable event by sending the information we own
    to the rid depending on whether the rid is local or remote. We cannot
    generate vnodeReachable events here because we don't know which vnodes rid
    has.
    """
    self.sendToVMan(rid)

  def handleRouterUnreachableEvent(self, rid, udata):
    """ Process the router unreachable event by deleting all the vnodes 
    residing on rid.

    There is a slight asymmetry between the propagation of vnode location info
    and its deletion. The propagation of exposed vnode locations is 
    accomplished in two stages. First, the info is transfered to the local
    neighbor from the remote neighbor. Then, it is transferred from the local
    neighbor to all other local vmans. The deletion of exposed vnode location
    info (like local vnode location info) is done in one stage. When the remote
    router becomes unreachable, all local vmans delete the vnode info residing
    on the unreachable router.

    vman_locs   - static list, [vman_hosting_router_id], storing ids of routers 
    vnodes      - lid -> PLVNode, storing all local vnodes.
    vnode_locs  - lid -> set(rid), storing the locations of all vnodes
    evnodes     - rrid -> set(vlid), storing vnodes that we are exposing
    ervnodes    - rrid -> set(vlid), storing the remote vnodes exposed to self

    """
    lostvnodes = set()
    # remove vnodes whose location is rid and remember which vnodes are now
    # unreachable
    for lid, ridset in self.vnode_locs.iteritems():
      if rid in ridset:
        ridset.remove(rid)
        if len(ridset) == 0:
          lostvnodes.add(lid)
    
    if not self.rtr.isLocal(rid) and rid in self.ervnodes:
      self.ervnodes[rid] = set()
      
    # send vnode unreachable callbacks
    for lid in lostvnodes:
      for (cb, udata) in self.vucb.itervalues():
        cb(lid, udata)

  def consumeRouterPacket(self, raw_pkt, src, type, udata): # [Requires Orpa]
    """ Process the received packet based on the packet type.
    """
    if self.rtr.opts.debug:
      #print "Vman of", self.rtr.id, "received pkt of type", type
      pass
    
    if type == PLIrpI.VMAN_CTRL_PKT_TYPE:
      if self.rtr.opts.collect_stats:
        self.rtr.stats.vnode_exp_msg_count += 1 
      ps = raw_pkt.split(":", 1)
      assert len(ps) == 2
      mlist = pickle.loads(ps[1])
      elems = ps[0].split()
      assert len(mlist) == len(elems)
      for ind, el in enumerate(elems):
        if el == "vnodes":
          # src sent us the set of its vlids, src has to be local
          assert self.rtr.isLocal(src)
          
          # get the new (expired) lids at the src and generate vnode 
          # (un)reachable events for each such lid
          updatelids = set(mlist[ind])
          curlids = set()
          for lid, ridset in self.vnode_locs.iteritems():
            if src in ridset:
              curlids.add(lid)

          # update vnode_locs by first deleting all the references to src
          # and then putting all the lids in the message as located at src
          for lid, ridset in self.vnode_locs.iteritems():
            if src in ridset:
              ridset.remove(src)
          for lid in mlist[ind]:
            if lid not in self.vnode_locs:
              self.vnode_locs[lid] = set()
            self.vnode_locs[lid].add(src)

          # send vnode reachable callbacks [Requires Orpa because assumes 
          # that there are not other instances of this vnode]
          for lid in updatelids.difference(curlids):
            for (cb, udata) in self.vrcb.itervalues():
              cb(lid, udata)

          # send vnode unreachable callbacks
          for lid in curlids.difference(updatelids):
            for (cb, udata) in self.vucb.itervalues():
              cb(lid, udata)

        elif el == "ervnodes": # [need to generate callbacks]
          # src sent us vnode lids exposed to it, src has to be local
          assert self.rtr.isLocal(src)
          # for each rrid in the received ervnodes.keys(), update the vnode_locs
          # as above
          for rrid in mlist[ind]:
            for lid, ridset in self.vnode_locs.iteritems():
              ridset.discard(rrid)
            for lid in mlist[ind][rrid]:
              if lid not in self.vnode_locs:
                self.vnode_locs[lid] = set()
              self.vnode_locs[lid].add(rrid)

        elif el == "evnodes":
          # src sent us the vnodes it is exposing to us, src has to be remote
          assert not self.rtr.isLocal(src)

          # get the new (expired) lids at the src and generate vnode 
          # (un)reachable events for each such lid
          updatelids = set(mlist[ind])
          curlids = set()
          for lid, ridset in self.vnode_locs.iteritems():
            if src in ridset:
              curlids.add(lid)

          # for each lid in the received evnodes set, update the ervnodes
          # and vnode_locs by deleting references to src and then inserting them
          self.ervnodes[src] = mlist[ind]
          for lid, ridset in self.vnode_locs.iteritems():
            ridset.discard(src)
          for lid in mlist[ind]:
            if lid not in self.vnode_locs:
              self.vnode_locs[lid] = set()
            self.vnode_locs[lid].add(src)

          # send vnode reachable callbacks [Requires Orpa because assumes 
          # that there are not other instances of this vnode]
          for lid in updatelids.difference(curlids):
            for (cb, udata) in self.vrcb.itervalues():
              cb(lid, udata)

          # send vnode unreachable callbacks
          for lid in curlids.difference(updatelids):
            for (cb, udata) in self.vucb.itervalues():
              cb(lid, udata)

        else:
          assert False, "unknown data in vman ctrl message"
    elif type == PLIrpI.DATA_PKT_TYPE:
      dmsg = pickle.loads(raw_pkt)
      
      # while the packet's nh is local vnode lid, give the pkt to it
      while True:
        nh = dmsg.nh
        if nh in self.vnodes:
          should_continue = self.vnodes[nh].processDataMsg(dmsg)
          if not should_continue:
            break
          if self.rtr.opts.show_data_pkts:
            print "Processesed pkt", dmsg.data, "at vnode", nh
        else:
          if nh in self.vnode_locs and len(self.vnode_locs[nh]) > 0:
            smsg = pickle.dumps(dmsg)
            dst = iter(self.vnode_locs[nh]).next()
            if self.rtr.opts.show_data_pkts:
              print "Sending", dmsg.data, "over the net to", nh
            self.rtr.irp.sendPacket(smsg, dst , PLIrpI.DATA_PKT_TYPE)
            break
          else:
            if self.rtr.opts.show_data_pkts:
              print "The next hop for pkt", dmsg.data, \
                  "in unreachable from", self.rtr.id
            break
           
    else:
      assert False, "Unknown packet type %d" % type

  def sendToAllVMans(self):
    for vman in self.vman_locs:
      if self.rtr.nls.isNeighborReachable(vman):
        self.sendToVMan(vman)
  
  def sendToVMan(self, rid):
    if self.rtr.isLocal(rid):
      if rid in self.vman_locs:
        self.sendToLocal(rid)
    else:
      self.sendToRemote(rid)

  def sendToLocal(self, rid):
    """ Send self.vnodes and self.ervnodes to local router rid.
    
    Returns:
      True if successful. False otherwise
    """
    assert self.rtr.isLocal(rid)
    pkt = pickle.dumps([self.vnodes.keys(), self.ervnodes], 0)
    pkt = "vnodes ervnodes:" + pkt
    return self.irp.sendPacket(pkt, rid, PLIrpI.VMAN_CTRL_PKT_TYPE)
    
  def sendToRemote(self, rid):
    """ Send self.vnodes and self.ervnodes to local router rid.
    
    Returns:
      True if successful
    """
    assert not self.rtr.isLocal(rid)
    assert self.rtr.nls.isNeighbor(rid)
    if rid in self.evnodes:
      pkt = pickle.dumps([self.evnodes[rid]], 0)
      pkt = "evnodes:" + pkt
      return self.rtr.irp.sendPacket(pkt, rid, PLIrpI.VMAN_CTRL_PKT_TYPE)

  def readConfig(self):
    """ Read the config file and construct vman_locs
    """

    fd = open(self.rtr.opts.conf, "r")
    rid = self.rtr.id
    section = 0
    
    for line in fd:
      line = line.strip()
      if len(line) == 0 or line[0] == '#':
        continue
      
      ps = line.split()
      if ps[0] == "section":
        if ps[1] == "vman":
          section = 1
        else:
          section = 0
        continue

      if section == 1:
        assert len(ps) == 2, "vman line \"%s\" is malformed" % line
        cur_id = (int(ps[0]), int(ps[1]))
        if cur_id != rid:
          self.vman_locs.append(cur_id)

        
    
