
class PLVNode:
  """ The Pathlet VNode
  
  VNodes have two types of IDs: full IDs and logical IDs. FullIDs are 3-tuples
  (as_id, router_id, vnode_id) that represent an instantiation of a logical 
  VNode (as_id, vnode_id) on router (as_id, router_id). Logical IDs are the
  identifiers of distinct logical VNodes that can be physically distributed.

  VNodes are meant to be very lightweight forwarding tables. They don't 
  construct/withdraw pathlets. They don't know about physical topology other 
  than the id of the router and AS they are instanciated on. They don't do any
  access control. Access control has to be done by some agent before giving 
  the packet to the vnode. Their forwarding
  entries are completely virtual. They provide interface to forward a data pkt,
  install/remove a forwarding entry. They can provide some optional interfaces
  like querying for the current fentry, deleting all entries containing a given
  vnode, etc. 

  Vnodes are our destinations, so they have the ability to "deliver a packet".

  Vnodes are essentially very dump objects that just do what they are told to 
  do. Given this understanding, it does not make much sense to ask a question
  "does a vnode know about different instances of itself on different routers?"
  The answer is "yes and no". "Yes" because they are able to distinquish 
  between itself and its twin because each vnode knows its full id. "No" because
  this ability does not in any way impact their functionality. They just do 
  whatever is in their table, which is constructed by an external agent.

  Full ids are also called "complete ids" and are abbreviated cid.

  Structure of the forwarding table:
    fes - Forwarding Entries dictionary mapping fid -> (push_fids, 
          next_hop_vnode), where push_fids is a list of fids to be prepended
          to the FIDseq of the packet. This list can be empty. next_hop_vnode
          is an lid.

  
  The key sentence that defines the pathlet operation is:
  A pathlet F -> [A, B, C, ..., W] is a promise that if a packet with top 
  FID equal to F comes to vnode A, this packet with the F poped off will
  appear at W.

  It makes no sense to construct a new pathlet completely out of existing
  pathlets (because the user can already do it). Thus, each pathlet has a
  structure [A, .., B [X], [Y],...] where A, .., B is some sequence of local 
  vnodes (vnodes belonging to the constructing AS) that cannot be obtained 
  using existing pathlets (not necessarily local) and [X], [Y] are existing
  pathlets constructed by other ASes. The sequence A, ..., B can contain local
  pathlets in it. To use a local pathlet like this, the node simply prepends
  the local pathlet fid before giving the packet to the first vnode in the 
  local pathlet. When the packet leaves that local pathlet this fid will be
  removed and the packet will be flowing on our pathlet with our fid.
  
  """

  def __init__(self, rtr, cid):
    """ Initialize the vnode. 
    
    All the parameters are fixed for the run of the program. 

    Params:
      rtr  - The rtr object
      cid  - Complete id of the self, (as_id, router_id, vnode_id).
      
    Return:
      None
    """

    self.rtr = rtr
    self.cid = cid
    self.lid = (cid[0], cid[2])
    self.fes = {}

  def getNumEntries(self):
    return len(self.fes)
    
  def processDataMsg(self, msg):
    """ Processes the data message msg.

    The PLDataMsg contains the next-hop vnode id field, which must be equal 
    to the lid of self. The method either delivers the msg if self is the 
    destination (identified by a lid), or changes the msg as appropriate and 
    gives it to the dispatcher to deliver to the next vnode (as specified by
    the field in the msg object), or drops the packet if it is malformed

    Params:
      msg - Instance of PLDataMsg

    Returns:
      True  - if the packet should be forwarded further by the dispatcher
      False - if the packet has be delivered or dropped
    """
    assert msg.nh == self.lid, "Next hop vnode is incorrect"

    print msg

    # dec ttl and drop if expired
    msg.ttl = msg.ttl - 1
    if msg.ttl <= 0:
      if self.rtr.opts.show_data_pkts:
        print "Pkt", msg.data, "is dropped because ttl expired"   
      return False
      
    # If self is the destination, deliver the packet.
    if msg.dst == self.lid:
      assert len(msg.fids) == 0, "Unexpected number of fids left at dst: " + \
          str(msg.fids)
      if self.rtr.opts.show_data_pkts:
         print "Pkt", msg.data, "is delivered"   
      return False

    assert len(msg.fids) > 0, "No fids left at a non-destination vnode"

    # Pop the first FID
    fid = msg.fids.pop()
    
    # If self knows about the fid, process the pkt. Else, drop it.
    if self.fes.has_key(fid):
      push_fids, nh = self.fes[fid]
      msg.fids.extend(push_fids)
      msg.nh = nh
      return True
    else:
      if self.rtr.opts.show_data_pkts:
        print "Vnode", self.lid, "dropped", msg.data, \
            "because it didn't have fid", fid   
      return False
      
  def installFE(self, fid, entry):
    """ Install the forwarding entry fid -> entry

    The implementation promises not to modify the list of fids in the entry, so
    these lists can be reused. If the entry already exists, overwrites it. In 
    debug mode, raises an error.

    Params:
      fid   - the FID to be installed.
      entry - the entry (push_fids, next_hop_vnode) to be installed 
              corresponding to the fid
      
    Returns:
      None
    """
    if self.rtr.opts.debug:
      assert not self.fes.has_key(fid), "FID %d already exists in %s as %s" % \
          (fid, str(self.cid), str(self.fes[fid]))

    self.fes[fid] = entry

  def removeFE(self, fid):
    """ Remove the entry corresponsing to fid and return it

    It is an error if the entry is not present.

    Params:
      fid - the FID whose entry is to be removed

    Returns:
      The removed entry
    """
    assert self.fes.has_key(fid), "vnode %s does not have fid %d" % \
        (str(self.cid), fid)

    entry = self.fes[fid]
    del self.fes[fid]
    return entry

  def containsFid(self, fid):
    return fid in self.fes


